surfSamps <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                     # posterior samples.
  x                  # The coordinate of the point where
                     # the surface should be evaluated.
){
  #######################################################################
  #
  # For internal use only. Get the posterior samples for the surface
  # height at a single location.
  #

  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  LogInfo  = dget(DatLG);                       # Misc stuff from log
  bdy  = LogInfo$etc$bdy;
  temps <- LogInfo$temps;
  nTemps <- length(temps);
  
  DatFL = paste(McmcFile,"Out.FL",sep="");     # Fixed-length data
  DatVL = paste(McmcFile,"Out.VL",sep="");     # Variable-length data
  if(!file.exists(DatFL))
    stop(paste("Unable to read fixed-length file",DatFL));
  if(!file.exists(DatVL))
    stop(paste("Unable to read vble-length file",DatVL));

  nSamps <- dim( getMcmc( McmcFile ) )[1]
  z = rep( 0, nSamps );
  
  o <- .C("surfSamps",
          DatFL = DatFL,                  DatVL = DatVL,
          xloc = as.double(x),
          z = as.double(z), nSamps = nSamps );
  
  return(o$z);
}

meanSurface <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                     # posterior samples.
  x,                 # The coordinates of the points where
                     # the mean surface should be evaluated.
  sampNum = -1       # 
){
  #######################################################################
  #
  # For internal use only. 
  #

  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  LogInfo  = dget(DatLG);                       # Misc stuff from log
  bdy  = LogInfo$etc$bdy;
  temps <- LogInfo$temps;
  nTemps <- length(temps);

  DatFL = paste(McmcFile,"Out.FL",sep="");     # Fixed-length data
  DatVL = paste(McmcFile,"Out.VL",sep="");     # Variable-length data
  if(!file.exists(DatFL))
    stop(paste("Unable to read fixed-length file",DatFL));
  if(!file.exists(DatVL))
    stop(paste("Unable to read vble-length file",DatVL));

  z = rep( 0, length( x ) );
  o <- .C("meanSurf",
          DatFL = DatFL,                  DatVL = DatVL,
          xlocs = as.double(x),
          temps = as.double(temps),       nTemps = as.integer( nTemps ),
          tempInd = as.integer( 0 ),
          z = as.double(z), nLocs = as.integer( length(x) ),
          sampNum = as.integer( sampNum - 1 ) );  # C indexes from zero
  
  return(o$z);
}

stdevSurface <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                     # posterior samples.
  x                  # The coordinates of the points where
                     # the SD surface should be evaluated.
){
  #######################################################################
  #
  # For internal use only. 
  #

  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  LogInfo  = dget(DatLG);                       # Misc stuff from log
  bdy  = LogInfo$etcs$bdy;
  temps <- LogInfo$temps;
  nTemps <- length(temps);
  
  DatFL = paste(McmcFile,"Out.FL",sep="");     # Fixed-length data
  DatVL = paste(McmcFile,"Out.VL",sep="");     # Variable-length data
  if(!file.exists(DatFL))
    stop(paste("Unable to read fixed-length file",DatFL));
  if(!file.exists(DatVL))
    stop(paste("Unable to read vble-length file",DatVL));

  z = rep( 0, length( x ) );
  o <- .C("stdevSurf",
          DatFL = DatFL,                  DatVL = DatVL,
          xlocs = as.double(x),
          temps = as.double(temps),       nTemps = as.integer( nTemps ),
          tempInd = as.integer( 0 ),
          z = as.double(z), nLocs = as.integer( length(x) ) );
  
  return(o$z);
}

percentAbove <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                     # posterior samples.
  x,                 # The coordinates of the points where
                     # the SD surface should be evaluated.
  cutoff             # The cutoff for which the percent of samples
                     # exceeding should be calculated
){
  #######################################################################
  #
  # For internal use only. 
  #

  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  LogInfo  = dget(DatLG);                       # Misc stuff from log
  bdy  = LogInfo$etc$bdy;

  DatFL = paste(McmcFile,"Out.FL",sep="");     # Fixed-length data
  DatVL = paste(McmcFile,"Out.VL",sep="");     # Variable-length data
  if(!file.exists(DatFL))
    stop(paste("Unable to read fixed-length file",DatFL));
  if(!file.exists(DatVL))
    stop(paste("Unable to read vble-length file",DatVL));

  z = rep( 0, length( x ) );
  o <- .C("percentAbove",
          DatFL = DatFL,                  DatVL = DatVL,
          xlocs = as.double(x),
          z = as.double(z), nLocs = as.integer( length(x) ),
          cutoff = as.double(cutoff) );
  
  return(o$z);
}

meanPlot <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                   # posterior samples.
  nxy=100,          # The number of grid points for the prediction maps.
  ps="",           # The name of the postscript file to generate.  If ps="",
                   # then the plot is send to the current device.  java.graph
                   # is the recommended device.
  xlim = NULL,
  ylim = NULL,
  xlab = "",
  ylab = "",
  cex = 2,
  lwd = 2,
  cex.axis = 0.5,
  cex.lab = 0.5,
  xRescale = 1     # factor by which to rescale the x-axis units
){
  
  #######################################################################
  #
  # Create a line plot of the posterior mean surface.
  # McmcFile is the base (root) name for the MCMC output files.
  # The MCMC files should be named <McmcFile>Out.FL and <McmcFile>Out.VL.  There
  # should also be a log file <McmcFile>Out.LG containing the log
  # information from the MCMC run.
  #

  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  MyId  = dget(DatLG);                       # Misc stuff from log
  bdy  = MyId$etc$bdy;
  if( is.null( xlim ) )
    xlim = bdy@xlim;

  lam = setGrid(bdy, nreq=nxy);
  nx = length(lam$x);
  xx = lam$x;

  lam$z <- meanSurface(McmcFile = McmcFile, x = xx);
  if( is.null( ylim ) )
    ylim = range( lam$z )
  
  if(nchar(ps)) {
    postscript(paste(ps,"-c.eps",sep=""),hori=TRUE, color.p=TRUE);
  }

  plot(x=lam$x*xRescale, y = lam$z, xlab=xlab,ylab=ylab,cex=cex, type = "l", lwd = lwd,
       xlim = xlim*xRescale, ylim = ylim, cex.axis = cex.axis, cex.lab = cex.lab);

  if(nchar(ps))
    dev.off();

  invisible(lam);
}

sampPlot <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                   # posterior samples.
  sampNum,          # The sample number
  nxy=100,          # The number of grid points for the prediction maps.
  ps="",           # The name of the postscript file to generate.  If ps="",
                   # then a graphical window is opened and no file is generated.
  xlim = NULL,
  ylim = NULL,
  xlab = "",
  ylab = "",
  lty  = 1,
  lwd = 2,
  cex.axis = 0.5,
  cex.lab = 0.5,
  cex = 1,
  xRescale = 1     # factor by which to rescale the x-axis units
){
  
  #######################################################################
  #  Create a line plot of the estimated function from a single sample
  #  of the Markov chain, given by sampNum.  McmcFile is the base (root)
  #  name for the MCMC output files.  The MCMC files should be named
  #  <McmcFile>Out.FL and <McmcFile>Out.VL.  There should also be a log
  #  file <McmcFile>Out.LG containing the log information from the MCMC
  #  run.

  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  MyId  = dget(DatLG);                       # Misc stuff from log
  bdy  = MyId$etc$bdy;
  if( is.null( xlim ) )
    xlim = bdy@xlim;

  lam = setGrid(bdy, nreq=nxy);
  nx = length(lam$x);
  xx = lam$x;

  lam$z <- meanSurface(McmcFile = McmcFile, x = xx, sampNum = sampNum);
  if( is.null( ylim ) )
    ylim = range( lam$z )
  
  if(nchar(ps)) {
    postscript(paste(ps,"-c.eps",sep=""),hori=TRUE, color.p=TRUE);
  }
  par(mar=c(5,5,4,2));

  plot(x=lam$x*xRescale, y = lam$z, xlab=xlab,ylab=ylab,cex=cex, type = "l",
       xlim = xlim*xRescale, ylim = ylim, lty = lty, lwd = lwd, cex.axis =
       cex.axis, cex.lab = cex.lab );

  if(nchar(ps))
    dev.off();

  invisible(lam);
}

stdevPlot <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                   # posterior samples.
  nxy=100,          # The number of grid points for the prediction.
  ps="",           # The name of the postscript file to generate.  If ps="",
                   # then a graphical window is opened and no file is generated.
  xlim = NULL,
  ylim = NULL,
  xlab = "",
  ylab = ""
){
  
  #######################################################################
  #
  # Create a line plot of the posterior mean surface.
  # McmcFile is the base (root) name for the MCMC output files.
  # The MCMC files should be named <McmcFile>Out.FL and <McmcFile>Out.VL.  There
  # should also be a log file <McmcFile>Out.LG containing the log
  # information from the MCMC run.
  #

  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  MyId  = dget(DatLG);                       # Misc stuff from log
  bdy  = MyId$etc$bdy;
  if( is.null( xlim ) )
    xlim = bdy@xlim;

  lam = setGrid(bdy, nreq=nxy);
  nx = length(lam$x);
  xx = lam$x;

  lam$z <- stdevSurface(McmcFile = McmcFile, x = xx);
  if( is.null( ylim ) )
    ylim = range( lam$z )

  if(nchar(ps)) {
    postscript(paste(ps,"-c.eps",sep=""),hori=TRUE, color.p=TRUE);
  }
  par(mar=c(5,5,4,2));

  plot(x=lam$x, y = lam$z, xlab=xlab,ylab=ylab,cex=2, type = "l", lwd = 2,
       xlim = xlim, ylim = ylim);

  if(nchar(ps))
    dev.off();

  invisible(lam);
}


larkPlot = function(
  McmcFile = "lrf",  # The base name for the files containing the
                   # posterior samples.
  ps="",           # The name of the postscript file to generate.  If ps="",
                   # then a graphical window is opened and no file is generated.
  sampNum=-1,      # Which MCMC sample number to plot.  If sampNum=-1
                   # then the last sample is plotted.
  xlim = NULL,     # The x-axis limits to be used.  If NULL, will use the
                   # boundary of the modeling region.
  ylim = NULL,     # The y-axis limits to be used.  If NULL, will use 0
                   # through the maximum kernel height.
  add = F,         # Whether to add the kernels to the current plot.
  meanLevel = 0,   # The line around which the kernels are drawn.
  nxy=100,          # The number of grid points for the prediction maps.
  altKernel = F,   # Whether or not to use the alternative kernel form.
  main = NULL,     # the title of the plot
  cex = 1,         # graphical parameters
  cex.axis = 1,
  cex.lab = 1,
  xlab = "",
  ylab = "",
  lwd = 1,
  xRescale = 1     # factor by which to rescale the x-axis units
){
  ###########################################################################
  # larkPlot
  #
  # Plot a single sample from the posterior distribution of the LARK model.
  # Draw each kernel on the plot.
  # 
  
  # Read output files from MCMCLoop
  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  DatFL = paste(McmcFile,"Out.FL",sep="");     # Fixed-length data
  DatVL = paste(McmcFile,"Out.VL",sep="");     # Variable-length data
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  if(!file.exists(DatFL))
    stop(paste("Unable to read fixed-length file",DatFL));
  if(!file.exists(DatVL))
    stop(paste("Unable to read vble-length file",DatVL));
  MyId  = dget(DatLG);                       # Misc stuff from log
  bdy  = MyId$etc$bdy;
  if( is.null( xlim ) )
    xlim = bdy@x
  temps <- MyId$temps
  nTemps <- length(temps)

  nAtoms <- 0
  o <- .C("getSampNatoms",
          DatFL = DatFL, DatVL = DatVL,
          temps = as.double(temps),       nTemps = as.integer( nTemps ),
          tempInd = as.integer(0),
          sampNum = as.integer(sampNum-1),                 # C indexes from 0
          nAtoms = nAtoms );  

  M <- o$nAtoms;

  xlocs <- rep(0, M);
  sig2s <- rep(0, M);
  heights <- rep(0, M);
  o <- .C("getLrfSamp",
          DatFL = DatFL,                  DatVL = DatVL,
          sampNum = as.integer(sampNum-1),                 # C indexes from 0
          temps = as.double(temps),       nTemps = as.integer( nTemps ),
          tempInd = as.integer(0),
          xlocs = as.double(xlocs), sig2s = as.double(sig2s),
          heights = as.double(heights), nAtoms = as.integer(M) );
  xlocs <- o$xlocs;
  sig2s <- o$sig2s;
  heights <- o$heights;
  
  if( is.null( ylim ) )
    ylim = c( 0, max(heights) )

  if( !add ){
    if(nchar(ps)) {
      postscript(paste(ps,"-d.eps",sep=""),
                 hori=TRUE, color.p=TRUE);
    }
    plot( x = xlim*xRescale, y = ylim, type = "n", cex=cex,
         cex.axis=cex.axis, cex.lab = cex.lab, xlab = xlab,
         ylab = ylab )
  }

  drawLocs <- seq( bdy@x[1], bdy@x[2], length.out = nxy )
  if( altKernel ){
    for( i in (1:M) ){
      kernDens <- abs( drawLocs - xlocs[i] );
      kernDens <- kernDens / sqrt( sig2s[i] );
      kernDens <- (1 + kernDens)^(-4);
      kernDens <- heights[i] * kernDens;
      kernDens <- kernDens + meanLevel;
      lines( drawLocs*xRescale,  kernDens, lwd=lwd )
    }
  } else {
    for( i in (1:M) ){
      kernDens <- ( drawLocs - xlocs[i] )^2;
      kernDens <- - kernDens / (2 * sig2s[i]);
      kernDens <- exp( kernDens );
      kernDens <- heights[i] * kernDens;
      kernDens <- kernDens + meanLevel;
      lines( drawLocs*xRescale,  kernDens, lwd=lwd )
    }
  }

  if( is.null(main) )
    title(paste("Mixture Components from MCMC Iteration #", sampNum));
  else
    title(main);
  if(nchar(ps)) dev.off();
  return();
}


getMcmc <- function(
  McmcFile = "lrf",  # The base name for the files containing the
                     # posterior samples.
  coda.loc = NULL,   # the directory for the coda library
  getZsums = F,      # Whether to sum the atom heights and return
                     # the resulting sums as a variable in the mcmc
                     # samples
  getLocSums = F,    # Whether to sum the atom locations and return
                     # the resulting sums as a variable in the mcmc
                     # samples
  getVarSums = F,    # Whether to sum the atom variances and return
                     # the resulting sums as a variable in the mcmc
                     # samples
  getAvgZ = F,       # Whether to average the atom heights and return
                     # the avg as a variable in the mcmc
                     # samples
  getAvgSig2 = F,    # Whether to obtain the average kernel variance
  getRandSig2 = F,   # Whether to obtain the variance of a single kernel,
                     # chosen at random, for each iteration of the chain.
  getRandZ = F       # Whether to obtain the magnitude of a single kernel,
                     # chosen at random, for each iteration of the chain.
  ){

  #######################################################################
  # getMcmc
  #
  # Creates an mcmc object (used by the coda package) containing
  # the posterior samples for M (the number of atoms in the levy process)
  # and the outcome precision.
  #

  # Read output files from MCMCLoop
  DatLG = paste(McmcFile,"Out.LG",sep="");     # Log file
  DatFL = paste(McmcFile,"Out.FL",sep="");     # Fixed-length data
  DatVL = paste(McmcFile,"Out.VL",sep="");     # Variable-length data
  if(!file.exists(DatLG))
    stop(paste("Unable to read lrf log file",DatLG));
  if(!file.exists(DatFL))
    stop(paste("Unable to read fixed-length file",DatFL));
  if(!file.exists(DatVL))
    stop(paste("Unable to read variable-length file",DatVL));

  MyId <- dget(DatLG);                          # Misc stuff from log
  temps <- MyId$temps;
  nTemps <- length(temps);
  fdat <- readBin(DatFL, what="double", n=1e8, size=8);   # Fixed-length data
  # calculate the number of parameters in the fixed-length file
  nParams <- 1 + nTemps*4    # one "parameter" is just the iteration number
  fdat <- matrix(fdat, ncol = nParams, byrow = T);
  nDat <- MyId$etc$N;                           # The # of data points
  nSamps <- dim(fdat)[1];                       # The # of posterior samples

  if( !is.element( "coda", search() ) ){
    if( is.null( coda.loc ) )
      library(coda);
    else {
      library(coda, lib.loc = coda.loc);
    }
  }
  
  mcmcDat <- matrix( nrow = nSamps, ncol = 0 )
  vars <- c()
  for( i in 1:nTemps ){
    thisTemp <- temps[i]

    thisTempInds <- (i-1)*4 + c( 2,4,5 )
    mcmcDat <- cbind( mcmcDat, fdat[, thisTempInds ] );
    if (nDat > 0){
      thisTempInd <- (i-1)*4 + 3
      rmse <- sqrt( fdat[,thisTempInd] / nDat );
    } else {
      rmse <- rep(0, nSamps);
    }

    mcmcDat <- cbind( mcmcDat, rmse );
    vars <- c( vars,
              paste("Temp", thisTemp, ":", c("M", "pre", "beta", "RMSE") ) )
  }
  if (getZsums){
    # get the zsums for the first temperature
    zsum <- rep( 0, nSamps );
    o <- .C("zsums",
      DatFL = DatFL,                  DatVL = DatVL,
      temps = as.double(temps),       nTemps = as.integer( nTemps ),
      tempInd = as.integer(0),        zsum = as.double(zsum) );  

    zsum = o$zsum;           # zsum is the sum of the jump heights for each
                             # iteration of the MCMC.
    print(length(zsum))
    mcmcDat <- cbind( mcmcDat, zsum );
    vars <- c( vars, "zsum" )
  }

  if (getAvgZ){
    # get the zsums for the first temperature
    zsum <- rep( 0, nSamps );
    o <- .C("zsums",
      DatFL = DatFL,                  DatVL = DatVL,
      temps = as.double(temps),       nTemps = as.integer( nTemps ),
      tempInd = as.integer(0),        zsum = as.double(zsum) );  

    zsum = o$zsum;           # zsum is the sum of the jump heights for each
                             # iteration of the MCMC.

    if(any( is.na( zsum ) ))
      stop("Some zsums are missing")
    zmean = zsum / mcmcDat[,1]
    zmean[ is.na( zmean ) ] = 0
    zmean[ is.inf( zmean ) ] = 0
    mcmcDat <- cbind( mcmcDat, zmean );
    vars <- c( vars, "zmean" )
  }

  if (getLocSums){
    locSum <- rep( 0, nSamps );
    o <- .C("locSums",
      DatFL = DatFL,                  DatVL = DatVL,
      temps = as.double(temps),       nTemps = as.integer( nTemps ),
      tempInd = as.integer(0),        
      locSum = as.double(locSum) );  

    locSum = o$locSum;       # locSum is the sum of the jump locations for each
                             # iteration of the MCMC.
    mcmcDat <- cbind( mcmcDat, locSum );
    vars <- c( vars,"locSum" )
  }

  if (getVarSums){
    varSum <- rep( 0, nSamps );
    o <- .C("varSums",
      DatFL = DatFL,                  DatVL = DatVL,
      temps = as.double(temps),       nTemps = as.integer( nTemps ),
      tempInd = as.integer(0),        
      varSum = as.double(varSum) );  

    varSum = o$varSum;       # varSum is the sum of the atom variances for each
                             # iteration of the MCMC.

    mcmcDat <- cbind( mcmcDat, varSum );
    vars <- c( vars, "varSum" )
  }

  if (getAvgSig2){
    avgSig2 <- rep( 0, nSamps );
    o <- .C("avgSig2",
      DatFL = DatFL,                  DatVL = DatVL,
      temps = as.double(temps),       nTemps = as.integer( nTemps ),
      tempInd = as.integer(0),        
      avgSig2 = as.double(avgSig2) );  

    avgSig2 = o$avgSig2;

    mcmcDat <- cbind( mcmcDat, avgSig2 );
    vars <- c( vars, "avgSig2" )
  }

  if (getRandSig2){
    randSig2 <- rep( 0, nSamps );
    o <- .C("randSig2",
      DatFL = DatFL,                  DatVL = DatVL,
      temps = as.double(temps),       nTemps = as.integer( nTemps ),
      tempInd = as.integer(0),        
      randSig2 = as.double(randSig2) );  

    randSig2 = o$randSig2;
    mcmcDat <- cbind( mcmcDat, randSig2 );
    vars <- c( vars, "randSig2" )
  }

  if (getRandZ){
    randZ <- rep( 0, nSamps );
    o <- .C("randZ",
      DatFL = DatFL,                  DatVL = DatVL,
      temps = as.double(temps),       nTemps = as.integer( nTemps ),
      tempInd = as.integer(0),        
      randZ = as.double(randZ) );  

    randZ = o$randZ;
    mcmcDat <- cbind( mcmcDat, randZ );
    vars <- c( vars, "randZ" )
  }
  
  colnames(mcmcDat) <- vars
  lrfmcmc <- mcmc(mcmcDat);
    
  lrfmcmc
}
