
getRegressMcmc <- function(
  McmcFile = "resultsRegress.bin",  # The base name for the files containing the
                     # posterior samples.
  coda.loc = NULL,   # the directory for the coda library
  nCoefs = 31        # The number of regression coefficients in the model
  ){

  #######################################################################
  # getRegressMcmc
  #
  # Creates an mcmc object (used by the coda package) containing
  # the posterior samples for the regression.
  #

  # Read output files
  if(!file.exists(McmcFile))
    stop(paste("Unable to read regression samples file",DatLG));

  fdat <- readBin(McmcFile, what="double", n=1e8, size=8);   # Fixed-length data
  # remove the first (2+9)*6 + 31 elements of the vector, which correspond to the domain
  # bounds and the knot locations for the 6 predictor variables, and the column
  # means that have been subtracted to improve mixing.
  fdat <- fdat[-(1:97)]  
  fdat <- matrix(fdat, ncol = (nCoefs+1), byrow = T);
  nSamps <- dim(fdat)[1];                       # The # of posterior samples

  if( !is.element( "coda", search() ) ){
    if( is.null( coda.loc ) )
      library(coda);
    else {
      library(coda, lib.loc = coda.loc);
    }
  }

  colnames(fdat) <- c("phi", paste("beta", (1:nCoefs - 1), sep="") )
  regmcmc <- mcmc(fdat);
    
  regmcmc
}


getRegressFns <- function(
  McmcFile = "resultsRegress.bin",  # The base name for the files containing the
                     # posterior samples.
  coda.loc = NULL,   # the directory for the coda library
  nCoefs = 31,       # The number of regression coefficients in the model
  nGrid = 100,       # The size of the grid on which to evaluate the regression function
  model = rep(1,6),  # Which model to restrict to when calculating the function estimates.
  postQuantiles = c(0.025, 0.975),
                     # The posterior quantiles desired for calculating credible bands
  predPenalty,
  useIntercept = T
){
  #### get the posterior samples of the spline regression coefficients
  regmcmc = getRegressMcmc( McmcFile, coda.loc, nCoefs )
  nSamps = dim(regmcmc)[1]

  #### get the posterior samples of the model
  res = getModelSamps( McmcFile, nCoefs, predPenalty )
  modelSamps = res$predInclude
  sampWeights = res$sampWeights
  
  #### restrict to the set of posterior samples of regression coefficients
  #### that are associated with the specified model.
  thisModel = matrix( model, nrow = nSamps, ncol = length(model), byrow = T )
  isThisModel = rowSums(modelSamps == thisModel) == length(model)
  nSampsThisModel = sum( isThisModel )
  print( paste("There are", nSampsThisModel, "samples from this model, out of", nSamps, "total samples." ) )

  #### Get the column means that have been subtracted to improve mixing of the Markov
  #### chain.
  if(!file.exists(McmcFile))
    stop(paste("Unable to read regression samples file",DatLG));
  fdat = readBin(McmcFile, what="double", n=1e8, size=8);   
  colSubtr = fdat[67:97]

  if( nSampsThisModel > 1 ){
    temp = regmcmc[isThisModel,]
    regmcmc = mcmc( temp )

    #### get the posterior means of the regression coefficients associated with this
    #### model
    #postMeans = summary(regmcmc)$statistics[3:32,1]
  
    #### get the knot locations.
    # Read output files
    if(!file.exists(McmcFile))
      stop(paste("Unable to read regression samples file",DatLG));
    fdat <- readBin(McmcFile, what="double", n=1e8, size=8);   # Fixed-length data

    # get the first 2*6 elements of the vector, which correspond to the domain
    # bounds for the 6 predictors.
    predBounds <- matrix( fdat[1:12], nrow = 6, byrow = T )
  
    # get the first 9*6 elements of the vector, which correspond to the knot
    # locations for the 6 predictor variables.
    knotLocs <- matrix( fdat[13:66], nrow = 6, byrow = T )

    #### obtain the posterior mean and interval estimates of the functions.
    nPreds = 6
    predFn = list()
    predLocs = list()
    predLower = list()
    predUpper = list()
    interceptSamps = as.vector( regmcmc[,2] )
    for( k in 1:nPreds ){
      # for each predictor
    
      # choose a grid of x values
      x = seq(knotLocs[k,3], knotLocs[k,7], length = nGrid)
      basis = spline.des( x=x, knots = knotLocs[k,], ord = 3)$design

      # start with the intercept samples
      if(useIntercept)
        res = matrix( interceptSamps, nrow = nSampsThisModel, ncol = length(x), byrow = F )
      else
        res = matrix( 0, nrow = nSampsThisModel, ncol = length(x), byrow = F )

      for( j in 1:5 ){
        # create a matrix of the weighted basis function values,
        # where the rows correspond to the posterior samples
        # and the columns correspond to the time points.
        thisPred = basis[,j] - colSubtr[(k-1)*5 + j + 1]
        res = res +
          as.matrix(regmcmc[,(k-1)*5 + j + 2], nrow = nSampsThisModel) %*% matrix( thisPred, nrow = 1 )
      }

      thisPredLower = as.vector( unlist( colQuantiles( res, postQuantiles[1] ) ) )
      thisPredUpper = as.vector( unlist( colQuantiles( res, postQuantiles[2] ) ) )
      thisPredFn = colMeans( res )
      predLower = c(predLower, list(thisPredLower) )
      predUpper = c(predUpper, list(thisPredUpper) )
      predFn = c(predFn, list(thisPredFn) )
      predLocs = c(predLocs, list( x ))
    }
  
    return(list( predFn = predFn, predLocs = predLocs, predBounds = predBounds,
              predLower = predLower, predUpper = predUpper ))
  } else {
    if( nSampsThisModel == 0 )
      print( "No samples from this model" )
    else
      print( "Only one sample from this model" )
    return( NULL )
  }
}

getModelSamps = function(
  McmcFile = "resultsRegress.bin",  # The base name for the files containing the
                     # posterior samples.
  nCoefs = 31,      # The number of regression coefficients in the model
  predPenalty
  ){
  #######################################################################
  # getModelSamps
  #
  # Get the weighted posterior samples of the model, as represented by 
  # a vector of ones.
  #

  #### get the posterior samples of the spline regression coefficients
  regMcmc = getRegressMcmc( McmcFile, nCoefs )
    
  #### Convert posterior samples matrix into matrix of posterior samples
  #### of the predictor inclusion indicators.
  nPreds = 6
  nSamps = niter( regMcmc )

  binMat = (regMcmc[,3:32] != 0)
  # convert to a matrix of 0s and 1s
  binMat = binMat + 0
#print(regMcmc[1:10,])
#  print(binMat[1:10,])
      
  predInclude = matrix( 0, nrow = nSamps, ncol = 0 )
  for( k in 1:nPreds ){
    startInd = (k-1)*5 + 1
    thisInclude = rowSums(binMat[,startInd:(startInd+4)])
    thisInclude = as.numeric( thisInclude > 0 )
    predInclude = cbind(predInclude, thisInclude)
  }
#  print(predInclude[1:10,])

  # calculate the weights associated with the samples
  predPenalty = rep(predPenalty, each = 5)
  predPenalty = matrix( predPenalty, ncol = 1 )
  negLogPenalty = -log( predPenalty )
#print(negLogPenalty)
  sampWeights = binMat %*% negLogPenalty
#print(sampWeights[1:100])
  sampWeights = sampWeights - max(sampWeights)
  sampWeights = exp( sampWeights )
#print(sampWeights[1:100])
 # print(min(sampWeights))
  return( list( predInclude = predInclude, sampWeights = sampWeights ) )
}

postModeModel <- function(
  McmcFile = "resultsRegress.bin",  # The base name for the files containing the
                     # posterior samples.
  nCoefs = 31,       # The number of regression coefficients in the model
  priorInMcmc,       # An indicator of whether the model prior probability
                     # was included when simulating the Markov chain (alternative
                     # is that a uniform probability over models was used,
                     # in which case we adjust the model probabilities here.)
  predPenalty
  ){

  #######################################################################
  # postModeModel
  #
  # Finds the most probable model a posteriori
  # 
  #

  res = getModelSamps( McmcFile, nCoefs, predPenalty )
  predInclude = res$predInclude
  sampWeights = res$sampWeights
  nSamps = dim(predInclude)[1]
  
  #### create a matrix of the distinct models occurring in this dataset,
  #### in the order of frequency. 
  models = as.matrix( unique( data.frame( predInclude ) ) )
  nModels = dim(models)[1]
  nPreds = 6
  modelCounts = rep(0, nModels)
  sampsEachModel = rep(0, nModels)
  for( i in 1:nModels ){
    a = matrix(data = as.vector(models[i,]), nrow=nSamps,
                ncol = nPreds, byrow = T)
#print(a[1:10,])
#      print(predInclude[1:10,])
    isThisModel = rowSums( a == predInclude ) == nPreds
#print(isThisModel[1:10])
    modelCounts[i] = sum( sampWeights[isThisModel] )
    sampsEachModel[i] = sum( isThisModel )
  }
#print("**********************")
#  print(modelCounts)
#  print(models)
#  print(sampsEachModel)
  modelCounts = modelCounts / sum( modelCounts )
  modelOrder = rev( order( modelCounts ) )
  modelCounts = modelCounts[modelOrder]
  sampsEachModel = sampsEachModel[modelOrder]
  models = models[modelOrder,]

#  print("*************************")
#  print(modelCounts)
#  print(models)
#  print(sampsEachModel)

  if( !priorInMcmc ){
    #### adjust the counts according to the model prior probabilities, which
    #### were not used in the sampling.
    for( i in 1:nModels ){
      thisModel = models[i,]
      priorProb = 1
      for( j in 1:6 )
        if( thisModel[j] <= 0 )
          # if this predictor is not included in the model,
          # the prior probability should be (2^5-1) times
          # higher than if it is included in the model.  This
          # is because there are (2^5 - 1) times more 
          # configurations that have this variable included
          # in the model than configurations that do not.
          priorProb = priorProb * (2^5-1)
      modelCounts[i] = modelCounts[i] * priorProb
    }
  }

  # reorder the models after the adjustment
  modelOrder = rev( order( modelCounts ) )
  modelCounts = modelCounts[modelOrder]
  sampsEachModel = sampsEachModel[modelOrder]
  models = models[modelOrder,]
  
  modelProbs = modelCounts / sum(modelCounts)
  
  #### Return the models and their counts
  return ( list( modelProbs = modelProbs, sampsEachModel = sampsEachModel, models = models ) )
}

margPredProbs <- function(
  McmcFile = "resultsRegress.bin",  # The base name for the files containing the
                     # posterior samples.
  nCoefs = 31,       # The number of regression coefficients in the model
  priorInMcmc,       # An indicator of whether the model prior probability
                     # was included when simulating the Markov chain (alternative
                     # is that a uniform probability over models was used,
                     # in which case we adjust the model probabilities here.)
  predPenalty
){

  #######################################################################
  # margPredProbs
  #
  # Finds the marginal posterior probability of the inclusion of each 
  # predictor.
  #

  res = postModeModel( McmcFile, nCoefs, priorInMcmc, predPenalty )

  models = res$models
  modelProbs = res$modelProbs
  nModels = length(modelProbs)

  margProbs = rep(0, 6)
  for( i in 1:nModels ){
    for( j in 1:6 )
      if( models[i,j] > 0 )
        margProbs[j] = margProbs[j] + modelProbs[i]
  }
  
  return(margProbs)
}
  
