
#include "HarkSamp.h"
#include <vector.h>
#include <queue.h>
#include "stringify.h"
#include "math.h"
#include "jama_qr.h"
#include "jama_lu.h"
#include "tnt_array2d_utils.h"
#include "moreArray2dUtils.h"
#include "randGen.h"


/* debugging info: 
   To print debugging info at runtime, uncomment the #define DEBUG_REGR line below.
*/

#define TNT_BOUNDS_CHECK
//#define DEBUG_REGR
#define NKNOTS 3

#define SQR(x) ((x)*(x))

using TNT::Array1D;
using TNT::Array2D;
using std::cout;
using std::endl;

extern "C"
{
  double S_STDCALL S_rpois( double lambda );
  double S_STDCALL S_rgamma( double s );
  double S_STDCALL S_runif( void );
  double S_STDCALL S_rexp( void );
}

HarkSamp::HarkSamp
( const Array1D<double> &response,    // the response vector
  char *rootFile,                     // the root name for the files
				      // containing the LARK MCMC samples.
  int nTemps,                         // The number of temps used in
				       // the original simulations.
  double *priorVar,                   // The vector of the entries of
				      // the diagonal prior covariance
				      // for eta.
  int priorType,                      // priorType: 0= use the unit
				      // information prior of DiMatteo
				      // et al (2001; Biometrika).  In
				      // this case the priorVar
				      // argument is ignored.
				      // 1= use informative prior
				      // distributions; do not adjust
				      // for different variances of
				      // the predictor variables 
				      // when specifying the prior
				      // variance of the regression
				      // coefficients.  2= Do adjust.
  bool priorInMcmc,                   // Whether or not to include the
                                      // model prior probabilities in 
                                      // the Markov chain (alternative 
                                      // is to use a uniform prior)
  const Array1D<double> &predPenalty){ // The fake prior odds for each
                                      // predictor indicator; the MCMC
                                      // is simulated with these prior odds
                                      // and they are removed for analysis
                                      // of the results.  Can be used to 
                                      // improve mixing.

  priorInMcmc_ = priorInMcmc;
  if( priorInMcmc_ )
    cout << "Using model prior probabilities in the Markov chain sampling" << endl;
  else
    cout << "Not using model prior probabilities in the Markov chain sampling" << endl;
  predPenalty_ = TNT::log( predPenalty ); 
  cout << "Predictor indicator fake log-odds:" << predPenalty_ << endl;
  response_ = response;
  nData_ = response.dim();
  rootFile_ = rootFile;
  nTemps_ = nTemps;
  priorType_ = priorType;
  // specify the total number of predictors in the regression model.
  // there is the intercept, plus six predictors that are represented
  // using a quadratic spline with NKNOTS knots.
  nPredsTot = 1 + 6*(NKNOTS+2);

  priorVar_ = Array1D<double> ( nPredsTot, 0.0 );
  int j;
  for( j=0; j<nPredsTot; j++ )
    priorVar_[j] = priorVar[j];

  // Calculate y'y; this quantity is used in the Bayesian regression.
  yTy = 0.0;
  for( j=0; j<nData_; j++ )
    yTy += pow(response_[j],2.0);

  #ifdef DEBUG_REGR
    cout << "yTy: " << yTy << endl; fflush(stdout);
  #endif

  // Initialize XTX and XTy, two quantities used in the regression;
  // these will be updated as the X matrix changes.
  XTX = Array2D<double> (nPredsTot, nPredsTot);
  XTy = Array1D<double> (nPredsTot);

  #ifdef DEBUG_REGR
    cout << "nTemps: " << nTemps_ << "  Prior variances for eta: " 
       << priorVar_ << endl;
  #endif

  // Create arrays for the number of tries and acceptances for each
  // move type 
  nTry = Array1D<int> ( nPredsTot, 0 );  
  nAcc = Array1D<int> ( nPredsTot, 0 ); 

  // create probability vectors that will be used in the Markov chain
  // updating.  The first is a vector of uniform probabilities for
  // each predictor variable.  The second is a vector of two equal
  // probabilities.
  unifProbs_ = Array1D<double> ( nPredsTot, 1.0 );
  unifProbs_[0] = 0.0;   // Do not allow the intercept to be removed from the model
//  for( j=1; j<16; j++ )
//    unifProbs_[j] = 0.0;  // Turn off updating for the regression coefs other than \gamma
//  for( j=21; j<31; j++ )
//    unifProbs_[j] = 0.0;  
  unifProbs_ = unifProbs_ / sum(unifProbs_);
  twoProbs_ = Array1D<double> ( 2, 0.5 );

}


/**************************************************************************
 **  get a single sample from a fixed-length file (a sample consists
 **  of nAtoms, SSE, pre, and beta)
 **/  
int HarkSamp::getSamp( FILE *FLfile, int *nAtoms, double *sse, double *pre,
  double *beta ){

  double tempDbl[4];

  int nRead = fread( tempDbl, sizeof(double), 4, FLfile );
  if( nRead != 4 ){
    MESSAGE "Fixed-length file terminated unexpectedly" ERROR;
  }
  
  (*nAtoms) = (int) tempDbl[0];
  (*sse) = tempDbl[1];
  (*pre) = tempDbl[2];
  (*beta) = tempDbl[3];

  return(nRead);
}


/**************************************************************************
 **  get the parameters of a single atom from a variable-length file 
 **/  
int HarkSamp::getAtom( FILE *VLfile, double *x, double *sig2, double *height){

  double tempDbl [3];

  int nRead = fread( tempDbl, sizeof(double), 3, VLfile );
  (*x) = tempDbl[0];
  (*sig2) = tempDbl[1];
  (*height) = tempDbl[2];

  return(nRead);
}

/**************************************************************************
 **  Open a file containing the 
 **  posterior samples for a LARK model.
 **/  
FILE *HarkSamp::openOneFile
( int ind, 
  char *suffix ){

  // construct the fixed-length filename
  std::string fileName = rootFile_;
  fileName += stringify( (ind+1) );
  fileName += "Out.";
  fileName += suffix;
  #ifdef DEBUG_REGR
    if( ind < 10 ){ 
      cout << "Filename " << ind << ": " << fileName << endl;
    } 
    fflush(stdout);
  #endif
  FILE *thisFile = fopen( fileName.c_str(), "r" );
  if( thisFile == NULL ){
    cout << "Accessed nonexistent file " << fileName << endl;
    MESSAGE "Accessed nonexistent file" ERROR;
  }
  return( thisFile );
}


/**************************************************************************
 **  Open the files containing the 
 **  posterior samples for the separate LARK models.
 **/  
Array1D<FILE *> HarkSamp::openFiles( char *suffix ){

  // Open the file associated with the LARK posterior for each signal.
  std::string fileName;
  Array1D<FILE *> files( nData_ );
  int i;

  for( i=0; i<nData_; i++ ){
    // construct the fixed-length filename
    fileName = rootFile_;
    fileName += stringify( (i+1) );
    fileName += "Out.";
    fileName += suffix;
    #ifdef DEBUG_REGR
      if( i < 10 ){ 
        cout << "Filename " << i << ": " << fileName << endl;
      } 
      fflush(stdout);
    #endif
    files[i] = fopen( fileName.c_str(), "r" );
    if( files[i] == NULL ){
      cout << "Accessed nonexistent file " << fileName << endl;
      MESSAGE "Accessed nonexistent file" ERROR;
    }
  }

  return(files);
}

/**************************************************************************
 **  Check that the files containing the 
 **  posterior samples for the separate LARK models all exist.
 **/  
void HarkSamp::checkFiles( char *suffix ){

  // Open the file associated with the LARK posterior for each signal.
  std::string fileName;
  FILE *myFile;
  int i;

  for( i=0; i<nData_; i++ ){
    // construct the fixed-length filename
    fileName = rootFile_;
    fileName += stringify( (i+1) );
    fileName += "Out.";
    fileName += suffix;
    #ifdef DEBUG_REGR
      if( i < 10 ){ 
        cout << "Filename " << i << ": " << fileName << endl;
      } 
      fflush(stdout);
    #endif
    myFile = fopen( fileName.c_str(), "r" );
    if( myFile == NULL ){
      cout << "Accessed nonexistent file " << fileName << endl;
      MESSAGE "Accessed nonexistent file" ERROR;
    }
    fclose( myFile );
  }

}

/**************************************************************************
 **  Determine the number of posterior samples.
 **/  
int HarkSamp::getNiter(){

  FILE *thisFile = openOneFile( 0, "FL" );

  int nIter=0, nAtoms, k;
  double fromFile, sse, pre, beta;

  while( fread( &fromFile, sizeof(double), 1, thisFile ) ){
    // The iteration number has just been read from the file.

    for( k=0; k<nTemps_; k++ ){
      // for each temperature
      // get the parameters for this tempered chain,
      // from the FL file
      getSamp( thisFile, &nAtoms, &sse, &pre, 
	       &beta );

    }
    nIter++;
  }
  fclose(thisFile);

  return( nIter );
}

/**************************************************************************
 **  Sample from the conditional distribution of \phi^2 given \eta.
 **  See e.g. p.50 of Gelman, Carlin, Stern, and Rubin (2004).
 **/  
void HarkSamp::sampleResidVar
( const Array2D<double> &designMat,   // the design matrix
  const Array1D<double> &eta,         // the regression coefs
  double &phi2 ){                     // for storing the new residual variance

  #ifdef DEBUG_REGR
    cout << "design matrix" << designMat << endl;
  #endif
  
  // Calculate SSE
  Array1D<double> sqrErr = matmult(designMat, eta);
  #ifdef DEBUG_REGR
    cout << "sqrErr" << sqrErr << endl;
  #endif
  sqrErr = response_ - sqrErr;
  #ifdef DEBUG_REGR
    cout << "sqrErr" << sqrErr << endl;
  #endif
  sqrErr = sqrErr * sqrErr;
  #ifdef DEBUG_REGR
    cout << "sqrErr" << sqrErr << endl;
  #endif
  double SSE = TNT::sum(sqrErr);
  #ifdef DEBUG_REGR
    cout << "SSE: " <<SSE << endl;
  #endif

  // Sample phi^2
  double phi2inv = S_rgamma( nData_ / 2.0 );
  phi2inv = 2 * phi2inv / SSE;
  phi2 = 1/phi2inv;
}

/**************************************************************************
 **  Sample from the conditional distribution of \eta given \phi^2.
 **  Computation is done as described in Gelman, Carlin, Stern, and
 **  Rubin (2004), Sec.14.2.
 **/  
void HarkSamp::sampleCoefs
( const Array2D<double> &designMat,   // the design matrix
  Array1D<double> &eta,               // for storing the regression coefs
  double phi2 ){                     // the residual variance

  #ifdef DEBUG_REGR
    cout << "design matrix" << designMat << endl;
  #endif

  // Obtain the QR factorization of designMat.
  JAMA::QR<double> designQR( designMat );
  Array2D<double> R = designQR.getR(),
    Q = designQR.getQ();

  #ifdef DEBUG_REGR
    cout << "Q: " << Q << endl << "R: " << R << endl;  fflush(stdout);
  #endif

  // calculate Q'y
  int i,j, nPreds = designMat.dim2();
  Array1D<double> QTy( nPreds, 0.0 );
  for( i=0; i<nPreds; i++ ){
    for( j=0; j<nData_; j++ )
      QTy[i] += Q[j][i] * response_[j];
  }

  // Obtain R^{-1}
  JAMA::QR<double> solveQR( R );
  Array2D<double> Rinv = solveQR.solve( TNT::identity( nPreds ) );
  
  // Calculate etaHat
  Array1D<double> etaHat = solveQR.solve( QTy );

  #ifdef DEBUG_REGR
    cout << "QTy: " << QTy << endl;
    cout << "responses" << response_ << endl;
    cout << "Rinv" << Rinv << endl;
    cout << "etaHat: " << etaHat[0] << " " << etaHat[1] <<
      " " << etaHat[2] << " " << etaHat[3] << " " << etaHat[4] <<
      " " << etaHat[5] << endl;
  #endif

  // Sample eta
  for( i=0; i<nPreds; i++ )
    eta[i] = S_rnorm();
  eta = matmult( Rinv, eta );
  eta = eta * pow(phi2, 0.5);
  eta = eta + etaHat;
}

/**************************************************************************
 **  Calculate the least squares estimate of the regression
 **  coefficient vector, and the MSE.  Computations are done as
 **  described in Gelman, Carlin, Stern, and Rubin (2004), p.355-357.
 **/
Array1D<double> HarkSamp::LSE
( const Array2D<double> &designMat,     // the design matrix
  double &MSE ){                        // Will be set to the MSE

  // Obtain the QR factorization of designMat.
  JAMA::QR<double> designQR( designMat );
  Array2D<double> R = designQR.getR(),
    Q = designQR.getQ();

  #ifdef DEBUG_REGR
    cout << "Q: " << Q << endl << "R: " << R << endl;  fflush(stdout);
  #endif

  int i,j, nPreds = designMat.dim2();

  // calculate Q'y
  Array1D<double> QTy( nPreds, 0.0 );
  for( i=0; i<nPreds; i++ ){
    for( j=0; j<nData_; j++ )
      QTy[i] += Q[j][i] * response_[j];
  }
  
  // Calculate etaHat
  JAMA::QR<double> solveQR( R );
  Array1D<double> etaHat = solveQR.solve( QTy );

  // calculate the SSE
  Array1D<double> sqrErr = matmult(designMat, etaHat);
  #ifdef DEBUG_REGR
    cout << "sqrErr" << sqrErr << endl;
  #endif
  sqrErr = response_ - sqrErr;
  #ifdef DEBUG_REGR
    cout << "sqrErr" << sqrErr << endl;
  #endif
  sqrErr = sqrErr * sqrErr;
  #ifdef DEBUG_REGR
    cout << "sqrErr" << sqrErr << endl;
  #endif
  double SSE = TNT::sum(sqrErr);
  #ifdef DEBUG_REGR
    cout << "SSE: " <<SSE << endl;
  #endif
  // Calculate the MSE
  MSE = SSE / (nData_ - nPredsTot);

  return( etaHat );
}

/**************************************************************************
 **  Update the model (predsInd) according to its marginal posterior probability
 **/  
bool HarkSamp::updatePredsInd
( const Array2D<double> &designMat,   // the design matrix
  Array1D<int> &predsInd,             // the model (to be updated)
  double &logMargProb,                // the value of the log marginal posterior
                                      // probability (will be updated)
  double priora,                      // priora,priorb: hyperparams for IG prior on phi2.
  double priorb,
  const Array2D<double> &priorPrec){  // After isolating the columns
				      // and rows specified by
				      // predsInd, this is the prior
				      // precision matrix for the
				      // regression coefficients.


  // Propose a new model by changing one of the values of the predsInd vector.
  // Choose the predictor index and the new value uniformly
  int toChange = rdiscrete( unifProbs_ ),
    newVal = rdiscrete( twoProbs_ ),
    //store the existing value so that can restore predsInd vector if
    //move is rejected
    oldVal = predsInd[toChange];

  if( newVal != oldVal ){
    predsInd[toChange] = newVal;

    double newLogProb = logModelProb( designMat, predsInd, priora, priorb,
				      priorPrec );
    
    // Calculate the log acceptance probability
    double logH = newLogProb - logMargProb;

    #ifdef DEBUG_REGR
      cout << "Proposing to update predsInd to " << predsInd << endl
	   << "New log prob: " << newLogProb 
	   << "   Log acceptance probability: " << logH << endl; fflush(stdout);
    #endif
    
    double samp = S_rexp();
    nTry[toChange]++;
    if( logH + samp > 0 ) {   // accept the proposed change
      logMargProb = newLogProb;     // update the log model prob.
      nAcc[toChange]++;
      #ifdef DEBUG_REGR
        printf("proposal accepted\n");
      #endif
      return (true);
    } else {
      // Reject the proposed addition.  Change predsInd back to its
      // previous value
      predsInd[toChange] = oldVal;

      #ifdef DEBUG_REGR
        printf("proposal rejected\n");
      #endif
      return (false);
    }
  }
  return(true);
}

/**************************************************************************
 **  Calculate the posterior density of a Bayesian linear regression
 **  model with conjugate prior.  The prior is a normal inverse gamma
 **  distribution.  The parameters of the prior inverse gamma density
 **  are given by priora and priorb.  The prior mean of the normal is
 **  taken to be zero, except for the intercept, which has prior mean
 **  eta0hat.  The prior precision matrix of the normal dist'n is
 **  specified by priorPrec.  The posterior density is also a normal
 **  inverse gamma, characterized by four quantities.  These are a, b:
 **  the params of the inverse gamma density; etaHat, Veta: the
 **  parameters of the normal density; note that the variance Veta
 **  should be multiplied by the value \phi^2 of the
 **  inverse-gamma-distributed variance parameter to get the posterior
 **  variance of eta.  Computations are done as described in Sudipto
 **  Banerjee's PUBH7440 course notes at www.biostat.umn.edu/~ph7440.
 **/
void HarkSamp::bayesRegression
( const Array2D<double> &designMat,   // the design matrix
  const Array1D<int> &predsInd,       // the vector of inclusion indicators
  double &a,                          // a,b: Will be set to the posterior values
  double &b,
  Array1D<double> &etaHat,            // Will be set to the posterior mean vector
  Array2D<double> &VetaInv,           // Will be set to the posterior
				      // precision matrix
  double priora,                      // priora,priorb: hyperparams for IG prior on phi2.
  double priorb,
  const Array2D<double> &priorPrec){  // After isolating the columns
				      // and rows specified by
				      // predsInd, this is the prior
				      // precision matrix for the
				      // regression coefficients.

  // create a matrix that contains only the relevant elements of the
  // prior precision matrix vector for eta
  Array2D<double> priorPrecUsed = subarray( priorPrec, predsInd,
			      predsInd );

  #ifdef DEBUG_REGR
    cout << "current prior precision matrix: " << priorPrecUsed << endl;
  #endif

  int i, nPreds = priorPrecUsed.dim2();

  // Obtain V_\eta^{-1}.  First obtain the relevant entries of X'X.
  VetaInv = subarray( XTX, predsInd, predsInd );

  #ifdef DEBUG_REGR
    cout << "VetaInv: " << VetaInv << endl; fflush(stdout);
  #endif

  // Add to it the prior precision matrix.
  VetaInv = VetaInv + priorPrecUsed;

  #ifdef DEBUG_REGR
    cout << "VetaInv: " << VetaInv << endl;  fflush(stdout);
  #endif

  // obtain the relevant entries of X'y
  Array1D<double> XTyAdj = subarray( XTy, predsInd );

  // Add to it the prior mean term
  for( i=0; i<nPreds; i++ )
    XTyAdj[i] += eta0hat * priorPrecUsed[i][0];

  #ifdef DEBUG_REGR
    cout << "XTyAdj: " << XTyAdj << endl; fflush(stdout);
  #endif

  // Calculate etaHat = Veta * XTy
  JAMA::QR<double> solveQR( VetaInv );
  etaHat = solveQR.solve( XTyAdj );

  #ifdef DEBUG_REGR
    cout << "etaHat: " << etaHat << endl; fflush(stdout);
  #endif

  // Calculate the scale parameter b of the inverse gamma posterior of
  // \phi^2.  First calculate the inner product of \eta-hat and X'y
  double etaHatXTy = 0.0;
  for( i=0; i<nPreds; i++ )
    etaHatXTy += etaHat[i] * XTyAdj[i];

  #ifdef DEBUG_REGR
    cout << "etaHatXTy: " << etaHatXTy << endl;  fflush(stdout);
  #endif

  // Subtract from y'y to get the scale param. of the IG dist'n.  Also
  // include the prior mean term.
  b = 0.5 * ( yTy - etaHatXTy );
  b += priorb + 0.5 * pow(eta0hat, 2.0) * priorPrecUsed[0][0];

  // Calculate the shape parameter a of the IG posterior for \phi^2.
  a = (double)nData_ / 2.0;
  a += priora;
}


/**************************************************************************
 **  Calculate the prior probability of the model given by predsInd,
 **  up to a normalizing constant.  Adjust according to predPenalty_.
 **/  
double HarkSamp::logModelPrior
( const Array1D<int> &predsInd ){

  double logPriorProb = 0.0;
  bool predIncluded;
  int j, k, startInd,
    nBasesPerPred = NKNOTS + 2;

  if( priorInMcmc_ ){
    for( j=0; j<6; j++ ){
      startInd = j*nBasesPerPred + 1;

      // check whether the jth predictor is included in the model
      // (i.e. whether any one of the coefficients associated with
      // this predictor is included in the model).
      predIncluded = false;
      for( k=startInd; k< startInd + nBasesPerPred; k++ )
	if( predsInd[k] > 0 )
	  predIncluded = true;

      // If this predictor is not included in the model, adjust the prior
      // probability upwards.
      if( !predIncluded )
	logPriorProb += log( pow(2.0, nBasesPerPred) - 1 );
    }
  }
  for( j=0; j<30; j++ ){
    if( predsInd[j+1] > 0 )
      logPriorProb += predPenalty_[j];
  }

  #ifdef DEBUG_REGR
    cout << "predsInd: " << predsInd << endl;
    cout << "log prior prob. " << logPriorProb << endl;  fflush(stdout);
  #endif
  return (logPriorProb);
}


/**************************************************************************
 **  Calculate the posterior probability of the model given by
 **  predsInd, up to a normalizing constant.  Computations done as
 **  described on p.349 of George and McCulloch (1997:Statistica
 **  Sinica).
 **/  
double HarkSamp::logModelProb
( const Array2D<double> &designMat,   // the design matrix
  const Array1D<int> &predsInd,
  double priora,                      // priora,priorb: hyperparams for IG prior on phi2.
  double priorb,
  const Array2D<double> &priorPrec ){       // After isolating the columns
				      // and rows specified by
				      // predsInd, this is the prior
				      // precision matrix for the
				      // regression coefficients.

  // Obtain the parameters of the posterior density of the regression model,
  // conditional on the specified set of predictors.
  Array2D<double> VetaInv;
  Array1D<double> etaHat;
  double a, b;

  double startSolve = tk.clockSecs();
  bayesRegression( designMat, predsInd, a, b, etaHat, VetaInv, priora, priorb,
		   priorPrec );

  // Calculate the determinant of the posterior precision VetaInv
  JAMA::LU<double> detCalc( VetaInv );
  double detVetaInv = detCalc.det();

  #ifdef DEBUG_REGR
    cout << "Determinant of VetaInv: " << detVetaInv << endl;
  #endif

  // Get the prior precision matrix.
  Array2D<double> priorPrecUsed = subarray( priorPrec, predsInd,
			      predsInd );

  // Calculate the determinant of the prior precision matrix 
  detCalc = JAMA::LU<double>( priorPrecUsed );
  double detPriorInv = detCalc.det();

  solveTime += tk.clockSecs() - startSolve;

  #ifdef DEBUG_REGR
    cout << "Prior precision Determinant: " << detPriorInv << endl;
  #endif

  // Calculate the marginal model posterior probability (up to a
  // normalizing constant).  Assumes a uniform prior over models.
  double logMargProb = logModelPrior( predsInd );
  logMargProb += 0.5*log(detPriorInv);
  logMargProb += -0.5*log(detVetaInv);
  logMargProb += -a * log(b);

  #ifdef DEBUG_REGR
    cout << "Log Prob. of proposed model: " << logMargProb << endl;
  #endif

  return( logMargProb );
}


/**************************************************************************
 **  Obtain a single sample from the posterior dist'n of the
 **  coefficients eta and the variance phi2 of a linear regression
 **  model, given the vector of inclusion indicators for the
 **  predictors.  The prior \pi( \phi^2 ) = IG( priora, priorb ) and
 **  \pi( \eta | \phi^2 ) = N( 0, \phi^2 priorPrec^{-1} ) is used,
 **  except that the prior mean for eta_0 is taken to be = eta0hat.
 **  Computations are done as described in Sudipto Banerjee's PUBH7440
 **  course notes at www.biostat.umn.edu/~ph7440
 **/  
void HarkSamp::sampleCoefsAndVar
( const Array2D<double> &designMat,   // the design matrix
  Array1D<double> &eta,               // for storing the regression coefs
  double &phi2,                       // for storing the residual variance
  const Array1D<int> &predsInd,
  double priora,                      // priora,priorb: hyperparams for IG prior on phi2.
  double priorb,
  const Array2D<double> &priorPrec){        // After isolating the columns
				      // and rows specified by
				      // predsInd, this is the prior
				      // precision matrix for the
				      // regression coefficients.

  // Obtain the parameters of the posterior density of the regression model,
  // conditional on the specified set of predictors.
  Array2D<double> VetaInv;
  Array1D<double> etaHat;
  double a, b;

  bayesRegression( designMat, predsInd, a, b, etaHat, VetaInv, priora, priorb,
		   priorPrec );

  #ifdef DEBUG_REGR
    cout << "Bayesian regression results:  a= " << a 
	 << "  b= " << b << "  etaHat= " << etaHat << endl
	 << "VetaInv = " << VetaInv << endl;
  #endif

  // Sample phi^2
  double phi2inv = S_rgamma( a );
  phi2inv = phi2inv / b;
  phi2 = 1/phi2inv;

  // Sample eta.  First compute the matrix Veta.
  JAMA::QR<double> solveQR( VetaInv );
  int nPreds = VetaInv.dim2();
  Array2D<double> Veta = solveQR.solve( TNT::identity(nPreds) );
  // Veta is supposed to be symmetric, but due to slight numerical
  // imprecision this may not be the case.  Make Veta symmetric.
  Veta = 0.5* (arrayTranspose(Veta) + Veta);

  #ifdef DEBUG_REGR
    cout << "Veta: " << Veta << endl;
  #endif

  Veta = Veta * phi2;
  Array1D<double> etaShort = rmvnorm( etaHat, Veta );

  #ifdef DEBUG_REGR
    cout << "etaShort " << etaShort << endl;
  #endif

  // The vector etaShort is the nonzero elements of eta.  Fill in the
  // zeros to obtain eta.
  eta = 0.0;
  int j=0;
  for( int i=0; i<nPredsTot; i++ ){
    if( predsInd[i] > 0 ){
      eta[i] = etaShort[j];
      j++;
    }
  }

  #ifdef DEBUG_REGR
    cout << "Sampled phi2 = " << phi2 << "  and eta = " << eta << endl;
  #endif
}


/**************************************************************************
 **  Perform several iterations of a Markov chain designed to sample
 **  from the posterior distribution of a regression model.
 **/  
void HarkSamp::regressSamp
( const Array2D<double> &designMat,   // the design matrix
  Array1D<double> &eta,               // for storing the regression coefs
  double &phi2,                       // for storing the residual variance
  Array1D<int> &predsInd,	      // for storing the predictor indicators
  double &logMargProb,                // the value of the log marginal posterior
                                        // probability of the current
                                        // model (will be updated).
  int nIter,                            // The number of iterations of the chain 
                                        // to run.
  double a,                             // a,b: hyperparams for IG prior on phi2.
  double b,
  const Array2D<double> &priorPrec){        // After isolating the columns
				      // and rows specified by
				      // predsInd, this is the prior
				      // precision matrix for the
				      // regression coefficients.

  for( int iter = 0; iter < nIter; iter++ ){

    // Update the predictor indicator vector from its marginal
    // posterior distribution.  Store the new log model probability
    // in logMargProb
    updatePredsInd( designMat, predsInd, logMargProb, a, b,
		    priorPrec );

  }

  // sample the coefficients and variance parameter conditional on
  // the predictor indicator vector
  sampleCoefsAndVar( designMat, eta, phi2, predsInd, a, b,
		     priorPrec );
}



/**************************************************************************
 **  Read in the design matrices from file, for each posterior sample.
 **/  

Array1D<Array2D<double> > HarkSamp::getDesignMats
(int nIter ){

  int i, j, k, l, res;
  
  // Create the 3D array that will hold the design matrix, for every 
  // posterior sample
  Array1D<Array2D<double> > allDesMat( nIter );
  for( i=0; i<nIter; i++ )
    allDesMat[i] = Array2D<double> ( nData_, 7, 1.0 );

  double fromFile, x, sig2, height, sse, pre, beta, avgSig2, 
    avgHeight, avgMu;
  int nAtoms;

  FILE *thisFL, *thisVL;

  for( j=0; j<nData_; j++ ){
    // for each subject

    //open the subject fixed-length and variable-length files
    thisFL = openOneFile( j, "FL" );
    thisVL = openOneFile( j, "VL" );

    for( i=0; i<nIter; i++ ){
      // for each posterior sample

      // read in the iteration number for this posterior sample
      res = fread( &fromFile, sizeof(double), 1, thisFL );
      if( res < 1 )
	MESSAGE "Sample file ended unexpectedly" ERROR;
    
      avgSig2 = 0;  avgHeight = 0;  avgMu = 0;
      for( k=0; k<nTemps_; k++ ){
	// for each temperature
	// get the parameters for this tempered chain,
	// from the jth FL file
	getSamp( thisFL, &nAtoms, &sse, &pre, 
	       &beta );

	// if this is the first temperature,
	// store beta, pre, and nAtoms
	if( k == 0 ){
	  // fill in the design matrix--leave the first column as an 
	  // intercept
	  allDesMat[i][j][1] = beta;
	  allDesMat[i][j][2] = 1/pre;  //this transformation can improve the fit of the linear model
	  allDesMat[i][j][3] = nAtoms;
	}
    
	for(l=0; l<nAtoms; l++){
	  // for each atom
	  // get the parameters of this atom from the variable-length file
	  if( getAtom( thisVL, &x, &sig2, &height ) < 3 ){
  	    MESSAGE "Data in variable length file does not match fixed length file." ERROR;
	  }
	  if( k == 0 ){
	    // Add to avgSig2 and avgHeight
	    avgSig2 += sig2;
	    avgMu += x;
	    avgHeight += fabs(height);
	  }
	}
	if( k == 0 ){
	  // get the average sig2_m and gamma_m values
	  if( nAtoms > 0 ){
    	    avgSig2 = avgSig2 / nAtoms;
	    avgMu = avgMu / nAtoms;
	    avgHeight = avgHeight / nAtoms;
	  } else {
	    avgSig2 = 0;
	    avgMu = 0;
	    avgHeight = 0;
	  }
	  // store these
	  allDesMat[i][j][4] = avgHeight; 
	  allDesMat[i][j][5] = avgMu;
	  allDesMat[i][j][6] = avgSig2;
	}
      }
    }
    fclose( thisFL );
    fclose( thisVL );
  }
  return(allDesMat);
}

// Center the columns of a matrix (i.e. subtract the column mean).
// Return the vector of column means.
// Do not center columns which have omit = 1, but return 0 for the 
// corresponding element in the vector of column means.
Array1D<double> colCenter( Array2D<double> &mat,
			   const Array1D<int> &omit){
  int i,j, n = mat.dim1(), m= mat.dim2();

  Array1D<double> colMeans( m );
  for( j=0; j<m; j++ ){
    colMeans[j] = 0.0;
    if( omit[j] != 1 ){
      for( i=0; i<n; i++ )
	colMeans[j] += mat[i][j];
      colMeans[j] /= n;
      for( i=0; i<n; i++ )
	mat[i][j] = mat[i][j] - colMeans[j];
    } else {
      for( i=0; i<n; i++ )
	mat[i][j] = mat[i][j];
    }
  }
  return(colMeans);
}

// Subtract the elements of colCenters from the columns of
// mat.
void colSubtract( Array2D<double> &mat,
		const Array1D<double> &colCenters){
  int i,j, n = mat.dim1(), m= mat.dim2();

  for( j=0; j<m; j++ )
    for( i=0; i<n; i++ )	
      mat[i][j] = mat[i][j] - colCenters[j];
}

// Given a covariance matrix, convert to a correlation matrix.
Array2D<double> cor( const Array2D<double> &cov ){
  int i,j,n=cov.dim2();
  Array2D<double> res( n, n );
  for( i=0; i<n; i++ ){
    for( j=0; j<n; j++ ){
      res[i][j] = cov[i][j] / (pow(cov[i][i], 0.5) * pow(cov[j][j], 0.5));
    }
  }
  return(res);
}

Array2D<double> colQuantiles( const Array2D<double> &mat, 
			      const Array1D<double> &quants ){

  int i,j, n=mat.dim1(), m=mat.dim2(), nQuants = quants.dim(),
    elem;
  priority_queue<double> myQueue;
  vector<double> myVec (n);

  Array2D<double> res( m, nQuants );
  for(j=0; j<m; j++){
    // for each column

    for(i=0; i<n; i++){
      // for each row, put the element into the priority queue
      myQueue.push( mat[i][j] );
    }
    for(i=0; i<n; i++){
      // put the sorted elements into the vector
      myVec[n-i-1] = myQueue.top();
      myQueue.pop();
    }// myQueue should now be empty

    for(i=0; i<nQuants; i++){
      // find the appropriate element of the sorted sequence
      elem = (int) (quants[i] * n);
      if( elem >= n )
	elem = n-1;
      if( elem < 0 )
	elem = 0;

      // calculate the quantile
      res[j][i] = myVec[elem];
    }
  }

  return( res );
}

/* Calculate XTX and XTy, two quantities that are used often in the
   regression calculation */
void HarkSamp::updateXTX
( const Array2D<double> &designMat ){

  double partsum;
  int i,j,k;
  for( i=0; i<nPredsTot; i++ )
    for( j=0; j<nPredsTot; j++ ){
      partsum = 0.0;
      for( k=0; k<nData_; k++ )
	partsum += designMat[k][i] * designMat[k][j];
      XTX[i][j] = partsum;
    }

  // calculate X'y
  for( i=0; i<nPredsTot; i++ ){
    partsum = 0.0;
    for( j=0; j<nData_; j++ )
      partsum += designMat[j][i] * response_[j];
    XTy[i] = partsum;
  }
}

/**************************************************************************
 **  Given a design matrix that has columns corresponding to the
 **  predictor variables, specify the knot locations for the
 **  variables.  We currently place knots at equally spaced quantiles
 **  of the variable.  Then we also include 3 knots at each boundary
 **  of the region.
 **/
Array2D<double>  fixKnotLocs
(const Array2D<double> &designMat){

  Array1D<double> quantiles( NKNOTS + 6 );
  quantiles[0] = 0.0;  quantiles[1] = 0.0;   quantiles[2] = 0.0;
  for( int i=0; i<NKNOTS; i++ ){
    quantiles[i+3] = (double)(i+1) / (NKNOTS + 1);
  }
  quantiles[NKNOTS + 3] = 1.0;  quantiles[NKNOTS+4] = 1.0;   quantiles[NKNOTS+5] = 1.0;
  cout << "quantiles: " << quantiles << endl;
  Array2D<double> knotLocs = colQuantiles( designMat, quantiles );

  // remove the intercept
  knotLocs = subarray( knotLocs,
		       1, knotLocs.dim1()-1,
		       0, knotLocs.dim2()-1 );

  return(knotLocs);
}

/**************************************************************************
 **  Given a design matrix that has columns corresponding to the
 **  predictor variables, calculate the design matrix corresponding to
 **  a B-spline basis of order 3 (quadratic) with knots at the
 **  locations given by knotLocs.  The rows of knotLocs correspond to
 **  the variables, while the columns correspond to the knots.  Here
 **  we assume the same number of knots for each predictor.
 **  Calculations are done according the recursive formulation at the
 **  top of p.90 of de Boor (2001).
 **/  
Array2D<double> bsplineDesign
(Array2D<double> &designMat,
 const Array2D<double> &knotLocs){

  int i,j,k1,n=designMat.dim1(), m=designMat.dim2(),
    nKnots = knotLocs.dim2();
  if( knotLocs.dim1() != m-1 ){
    cerr << "Number of variables in knot specification matrix must be same as # variables in design matrix." 
	 << endl;
    exit(1);
  }

  // It is possible for the values in designMat to be outside the
  // function domains defined by knotLocs.  If so, map them to the
  // boundary of the domain.
  for( i=0; i<n; i++ ){
    // for each subject

    for( j=0; j<(m-1); j++ ){
      // for each predictor variable 

      if( designMat[i][j+1] < knotLocs[j][0] )
	designMat[i][j+1] = knotLocs[j][0];
      else if( designMat[i][j+1] > knotLocs[j][nKnots-1] )
	designMat[i][j+1] = knotLocs[j][nKnots-1];
    }
  }

  #ifdef DEBUG_REGR
    cout << "Design matrix vals mapped to interior of domain " << endl;  
    fflush(stdout);
  #endif

  // Calculation is done according to de Boor's recursive formulation.
  // First calculate the bases of order 1.  
  double thisVal, thisKnot, otherKnot, thisWeight, temp;
  int thisIndex;
  Array2D<double> order1( n, (m-1)*(nKnots-1), 0.0 );
  for( i=0; i<n; i++ ){
    // for each subject

    for( j=0; j<(m-1); j++ ){
      // for each predictor variable 

      // get the predictor value
      thisVal = designMat[i][j+1];

      for( k1=0; k1<(nKnots-1); k1++ ){
	// for each relevant order-one basis

	// calculate the indicator of whether thisVal is in the interval
	// between the current knot and the next knot to the right
	if( (thisVal >= knotLocs[j][k1]) && (thisVal < knotLocs[j][k1+1]) ){
	  thisIndex = j*(nKnots-1) + k1;
	  order1[i][thisIndex] = 1.0;
	}
      }
    }
  }
  #ifdef DEBUG_REGR
    cout << "Order 1 b-spline bases: " << order1 << endl;  fflush(stdout);
  #endif

  // Now calculate the bases of order 2.  There is one fewer of them
  // for each predictor, than the bases of order 1.
  Array2D<double> order2( n, (m-1)*(nKnots-2), 0.0 );
  for( i=0; i<n; i++ ){
    // for each subject

    for( j=0; j<(m-1); j++ ){
      // for each predictor variable 

      // get the predictor value
      thisVal = designMat[i][j+1];

      for( k1=0; k1<(nKnots-2); k1++ ){
	// for each relevant basis

	// calculate the first term
	thisKnot = knotLocs[j][k1];
	otherKnot = knotLocs[j][k1+1];
	thisIndex = j*(nKnots-1) + k1;
	if( order1[i][thisIndex] > 0 ){
	  thisWeight = (thisVal - thisKnot) / (otherKnot - thisKnot);
	  temp = thisWeight * order1[i][thisIndex];
	} else {
	  temp = 0.0;
	}

	// calculate the second term
	thisKnot = knotLocs[j][k1+1];
	otherKnot = knotLocs[j][k1+2];
	thisIndex = j*(nKnots-1) + k1 + 1;
	if( order1[i][thisIndex] > 0 ){
	  thisWeight = (thisVal - thisKnot) / (otherKnot - thisKnot);
	  temp += (1-thisWeight) * order1[i][thisIndex];
	}

	// record the result
	thisIndex = j*(nKnots-2) + k1;
	order2[i][thisIndex] = temp;
      }
    }
  }
  #ifdef DEBUG_REGR
    cout << "Order 2 b-spline bases: " << order2 << endl;  fflush(stdout);
  #endif

  // Now calculate the bases of order 3.  There is one fewer of them
  // for each predictor, than the bases of order 2.  Also, we do not
  // use the last order-3 basis for each predictor: we instead
  // separately include an intercept term.  This ensures that the
  // resulting bases are linearly independent and span the function
  // space.
  Array2D<double> order3( n, 1 + (m-1)*(nKnots-4), 1.0 );
  for( i=0; i<n; i++ ){
    // for each subject

    for( j=0; j<(m-1); j++ ){
      // for each predictor variable 
      thisVal = designMat[i][j+1];

      for( k1=0; k1<(nKnots-4); k1++ ){
	// for each relevant basis

	// calculate the first term
	thisKnot = knotLocs[j][k1];
	otherKnot = knotLocs[j][k1+2];
	thisIndex = j*(nKnots-2) + k1;
	if( order2[i][thisIndex] > 0 ){
	  thisWeight = (thisVal - thisKnot) / (otherKnot - thisKnot);
	  temp = thisWeight * order2[i][thisIndex];
	} else {
	  temp = 0.0;
	}

	// calculate the second term
	thisKnot = knotLocs[j][k1+1];
	otherKnot = knotLocs[j][k1+3];
	thisIndex = j*(nKnots-2) + k1 + 1;
	if( order2[i][thisIndex] > 0 ){
	  thisWeight = (thisVal - thisKnot) / (otherKnot - thisKnot);
	  temp += (1-thisWeight) * order2[i][thisIndex];
	}

	// record the result
	thisIndex = 1 + j*(nKnots-4) + k1;
	order3[i][thisIndex] = temp;
      }
    }
  }

  return(order3);
}



/**************************************************************************
 **  Generate the samples from the linear regression posterior,
 **  given the samples from the LARK posterior distributions.
 **  Regression samples are saved to file.
 **/  
void HarkSamp::sample
(char *outFile,
 const Array1D<int> &initPredsInd,  // initial vector for the inclusion indicators
 int nIter2ndStage ){  

  int i, j, k1;
  solveTime = 0.0;

  int nIter = getNiter();  // determine the number of posterior samples

  // Read in the design matrices, for each posterior sample, from file.
  Array1D<Array2D<double> > allDesMat = getDesignMats( nIter );

  // Specify the prior hyperparameters.  To do this, first compute the
  // posterior mean of the predictor vectors for each subject.
  Array2D<double> postMean( nData_, 7, 1.0 ),
    postDesign;

  for( k1=1; k1<7; k1++ ){
    // for each predictor (leave out the intercept)
    for( j=0; j<nData_; j++ ){
      // for each observation, calculate the posterior mean estimate
      // of the predictor.
      postMean[j][k1] = 0.0; 
      for( i=0; i<nIter; i++ ){
	postMean[j][k1] += allDesMat[i][j][k1];
      }
      postMean[j][k1] = postMean[j][k1] / (double)nIter;
    }
  }

  // Decide on knot locations.  
  Array2D<double> knotLocs = fixKnotLocs( postMean );
  cout << "Knot locations: " << knotLocs << endl;

  // Calculate the .025 & .975 quantiles of the posterior mean predictor values.
  // These are the practical boundaries of the 
  // predictor domains.
  Array1D<double> quantiles (2);  quantiles[0] = .025;  quantiles[1] = .975;
  Array2D<double> predBounds = colQuantiles( postMean, quantiles );
  cout << "Predictor domains: " << predBounds << endl;  

  // convert to a design matrix for the nonparametric model
  postDesign = bsplineDesign( postMean, knotLocs );
  #ifdef DEBUG_REGR
    cout << "B-spline design matrix for posterior mean: " << postDesign << endl;
  #endif


  // If # obs >= # paramaters, find the LSE estimate of the intercept.
  // This will be used as the prior mean for the intercept.  Also
  // obtain the unbiased estimate of the residual variance.
  double MSE;
  if( nData_ > postDesign.dim2() ){
    Array1D<double> etaHat = LSE( postDesign, MSE );
    eta0hat = etaHat[0];
    cout << "Prior mean for the intercept: " << eta0hat << "  MSE: " << MSE << endl;  
    fflush(stdout);
  } else {
    eta0hat = 0.0;
  }


  // Choose the prior hyperparameters for the IG prior on phi^2.  If #
  // obs >= # paramaters, these are chosen so that the prior mean of
  // phi^{-2} is 1/MSE, and so that the variance of the response
  // variable is approximately one SD for phi^2 above the MSE.
  // Otherwise, we take a = b = 0.
  double a,b;
//  if( nData_ > postDesign.dim2() ){
//    double S2Y = var( response_ );
//    a = S2Y / MSE - 1;
//    a = pow(a, -2.0) + 2;
//    b = a * MSE;
//  }else {
    a = 0.0;  b = 0.0;
    //  }

  cout << "Prior hyperparams for phi2: a = " << a << "  b= " << b << endl;

  // Center the columns of the design matrix postMean (except for
  // the first column), so that XTX is the empirical covariance.
  // Store the column means, which will be subtracted from the 
  // design matrix in every iteration.
  Array1D<int> omit( nPredsTot, 0 );
  omit[0] = 1;
  Array1D<double> colMeans = colCenter( postDesign, omit );

  // Calculate the empirical covariance matrix associated with X
  Array2D<double> empCov = matmult( arrayTranspose( postDesign ), postDesign );
  empCov = empCov / (double)nData_;
  cout << "Empirical predictor covariance matrix:" <<
    empCov << endl;
  cout << "Correlation matrix: " << cor( empCov ) << endl;

 
  // Specify the prior precision matrix for eta.
  Array2D<double> priorPrec;
  if( priorType_ == 0 ){
    // unit information prior case
    // Take the prior precision to be the
    // empirical covariance of the predictors
    priorPrec = empCov;
    // Adjust the prior precision for the intercept.
    priorPrec[0][0] = priorPrec[0][0] / priorVar_[0];

  } else {
    // Unrecognized specification
    std::cerr << "Prior type not valid" << endl;
    exit(1);
  }
  cout << "Prior precision matrix: " << priorPrec << endl;
  // We have finished specifying prior hyperparameters. 

  // Create the regression coefficient vector object
  Array1D<double> eta( nPredsTot, 0.0 );

  // Initialize the predictor indicator vector
  Array1D<int> predsInd( initPredsInd );
  cout << "Initial predictor indicator vector: " << initPredsInd << endl;

  // Create design matrix objects
  Array2D<double> predsMat, designMat;

  if( nData_ <= nPredsTot ){
    MESSAGE "# of observations must be > # of predictors" ERROR;
  }

  // Open the file where the regression posterior samples
  // will be stored.
  FILE *regrFile = fopen( outFile, "w" );
  cout << "nData: " << nData_ << "   outfile: " << outFile << endl;  fflush(stdout);

  // Write the .025 & .975 quantiles of the posterior mean predictor
  // values to file.  Skip the first row, which corresponds to the intercept.
  for( i=1; i<predBounds.dim1(); i++ )
    for( j=0; j<predBounds.dim2(); j++ )
      fwrite( &(predBounds[i][j]), sizeof(double), 1, regrFile );

  // Write the knot locations to file
  for( i=0; i<knotLocs.dim1(); i++ )
    for( j=0; j<knotLocs.dim2(); j++ )
      fwrite( &(knotLocs[i][j]), sizeof(double), 1, regrFile );
  // Write the column means to file
  for( i=0; i<colMeans.dim1(); i++ )
    fwrite( &(colMeans[i]), sizeof(double), 1, regrFile );

  // initialize the variance parameters
  double phi2=1.0;

  double startTime = tk.clockSecs(),
    logMargProb = -DBL_MAX;  // force acceptance of the first proposal.
  cout << "nIter: " << nIter << "  nIter2ndStage: " << nIter2ndStage << endl;

  for( i=0; i<nIter; i++ ){
    // for each posterior sample

    // get the design matrix
    predsMat = allDesMat[i];
    // convert to a design matrix for the nonparametric model
    designMat = bsplineDesign( predsMat, knotLocs );
    // subtract the set of column centering values, for better Markov
    // chain mixing.
    colSubtract( designMat, colMeans );

    // calculate XTX and XTy, two arrays that are used often in the
    // regression computation
    updateXTX( designMat );

    // Simulate several iterations of a Markov chain designed to
    // sample from the posterior dist'n of a linear model (including
    // model averaging), conditional on this design matrix
    for( k1=0; k1<5; k1++ ){
      regressSamp( designMat, eta, phi2, predsInd, logMargProb,
		 nIter2ndStage, a, b, priorPrec );

      // write parameters to file
      fwrite( &phi2, sizeof(double), 1, regrFile );

      for( j=0; j<nPredsTot; j++ )
	fwrite( &(eta[j]), sizeof(double), 1, regrFile );
    }

    #ifdef DEBUG_REGR
      cout << "Regression estimates: " << phi2 << " " << eta << endl;
    #endif
  }

  cout << "Time spent on designated computations: " << solveTime
       << endl << "Total computation time: " << tk.clockSecs() - startTime << endl;
  cout << "Number of Tries: " << nTry << endl << "Number of Accepts: " << nAcc << endl;

   
  // close the file containing the regression parameter samples
  fclose( regrFile );

}
