#ifndef LEVY_INFER_H
#define LEVY_INFER_H



#include "tnt_array1d.h"
#include "tnt_array1d_utils.h"
#include "moreArray1dUtils.h"
#include <S.h>

using TNT::Array1D;


class LevyInfer{
 private:
  char *DatFL_, *DatVL_;

  Array1D<double> temps_;
  int nTemps;

  /**************************************************************************
   **  Get the height of a kernel centered at x, with variance = sig2
   **  and height= height, at the specified locations.
   **/  
  Array1D<double> kernelHeight(
    const Array1D<double> &xlocs, 
    double x, double sig2, double height );

  double kernelHeight(
    double xloc, 
    double x, double sig2, double height );
  
 public:
  LevyInfer( char *DatFL, char *DatVL, const Array1D<double> &temps );

  ~LevyInfer(){};

  /**************************************************************************
   **  Get the value of the posterior surface at a vector of
   **  locations, given by their coordinates, for the sample with
   **  index sampNum. 
   **/  
  Array1D<double> sampSurf( const Array1D<double> &xlocs, 
			    int sampNum, int tempInd );


  /**************************************************************************
   **  Get the value of the posterior surface at a particular
   **  location, for each posterior sample.
   **/  
  Array1D<double> surfSamps( double xloc, int tempInd );

  /**************************************************************************
   **  Get the value of the posterior surface for tempered chain
   **  tempInd, at a vector of locations given by their
   **  coordinates. If sampNum < 0 then the mean of the surface over
   **  the samples is returned, and otherwise the surface is evaluated
   **  for the sample with index sampNum (in the latter case a call is
   **  made to sampSurf).
   **/  
  Array1D<double> meanVals( const Array1D<double> &xlocs, 
			    int tempInd, int sampNum = -1 );

  /**************************************************************************
   **  Get the standard deviation of the posterior surface at a vector of
   **  locations, given by their coordinates. 
   **/  
  Array1D<double> stdevSurf( const Array1D<double> &xlocs, int tempInd );

  /**************************************************************************
   **  Get the percent of the posterior surface samples that is above
   **  a particular threshold, at a vector of locations given by their
   **  coordinates.
   **/  
  Array1D<double> percentAbove( const Array1D<double> &xlocs, double cutoff,
				int tempInd );

  /**************************************************************************
   **  Get a quantile of the posterior surface at a vector of
   **  locations, given by their coordinates. Not yet implemented.
   **/  
  Array1D<double> quantileSurf( const Array1D<double> &xlocs, 
				double quantile ) 
    { MESSAGE "Not yet implemented" ERROR; return(Array1D<double>()); };

  /**************************************************************************
   **  Get the sums of the atom heights for tempered chain tempInd,
   **  for each iteration of the MCMC.
   **/  
  Array1D<double> zsums( int tempInd );

  /**************************************************************************
   **  Get the sums of the atom locations for each iteration of the MCMC.
   **/  
  Array1D<double> locSums( int tempInd );

  /**************************************************************************
   **  Get the sums of the atom variances for each iteration of the MCMC.
   **/  
  Array1D<double> varSums( int tempInd );

  /**************************************************************************
   **  Get the average kernel variance (sigma^2) for each iteration of
   **  the MCMC.
   **/  
  Array1D<double> avgSig2( int tempInd );

  /**************************************************************************
   **  Get the variance (sigma^2) of a single kernel chosen at random,
   **  for each iteration of the MCMC.
   **/  
  Array1D<double> randSig2( int tempInd );

  /**************************************************************************
   **  Get the magnitude (z) of a single kernel chosen at random,
   **  for each iteration of the MCMC.
   **/  
  Array1D<double> randZ( int tempInd );

  /**************************************************************************
   **  Get the lrf for sample number sampNum.  The Array1D input
   **  objects must already have the correct number of elements (equal
   **  to the number of atoms in the sampNum iteration of the Markov
   **  chain).  To obtain this number of atoms, call the getSampNatoms
   **  function.
   **/  
  void getLrfSamp( int sampNum, Array1D<double> &xlocs, 
		   Array1D<double> &sig2, 
		   Array1D<double> &heights, 
		   int tempInd );

  /**************************************************************************
   **  Get the number of atoms in the sampNum iteration of the Markov chain.
   **/  
  int getSampNatoms( int sampNum, int tempInd );

};

#endif
