
#include "LevyMCMC.h"
#include "randGen.h"

/* debugging info: 
   To print high-level debugging info at runtime, add "-DDEBUG1" to the first line of the Makevars
   file.  To print more detailed debugging info, also add "-DDEBUG2" and possibly "-DDEBUG3" 
*/

// Define indexes for the move types
#define NmoveTypes 9
#define MV_PRE 0    // precision update
#define MV_BETA 1    // beta update
#define MV_NEWPT 2    // add an atom to the levy process
#define MV_MVLOC 3     // move an atom in the levy process
#define MV_MVMAG 4     // move an atom in the levy process
#define MV_CHSC 5     // change the scale of an atom in the levy process
#define MV_DELPT 6    // delete an atom in the levy process
#define MV_SPLIT 7    // delete an atom in the levy process
#define MV_MERGE 8    // delete an atom in the levy process

// Define indexes for the move types 
// that require a scale value for the proposal
#define Nscales 5
#define SCA_DRAWPT 0   // add an atom to the levy process
#define SCA_MVMAG 1    // change an atom's magnitude
#define SCA_CHSC 2    // change an atom's scale parameter
#define SCA_MVLOC 3    // change an atom's location
#define SCA_SPL 4    // step size for atom magnitude in a split move

#define SQR(x) ((x)*(x))

using TNT::Array1D;
using TNT::Array2D;
using std::cout;
using std::endl;

extern "C"
{
  double S_STDCALL S_rpois( double lambda );
  double S_STDCALL S_rgamma( double s );
  double S_STDCALL S_runif( void );
  double S_STDCALL S_rexp( void );
}


void printAcceptProb( double logProb ){
  if (logProb > 1000)
    printf("log accept prob = infty \n" );
  else if (logProb < -1000)
    printf("log accept prob = - infty \n" );
  else 
    printf("log accept prob = %f\n", logProb );
}

LevyMCMC::LevyMCMC(
     double *datX,                    // The data (x, z for each of datLen samples, where x is the location
     double *datZ,   int datLen,   // and z is the measurement)
     char *OutLG,                     // The names of the log file
     char *OutFL, char *OutVL,     // The names of the fixed-length and variable-length files to bes written
     double *xlim,                    // The bounding region within which to fit the surface.  xlim and ylim should 
                                      //be length-2 vectors
     int ItnMax,     int ItnBrn,  // ItnMax is the number of total iterations after burnin, while ItnBrn is 
                                      // the length of the burnin
     int SavMax,     double *pX,    // SavMax is the maximum number of saved iterations.  The resulting thinning 
                                      // is approximately 
                                      // ItnMax / SavMax.  pX is a vector of the probabilities of proposing a move 
                                      // for each of the parameters:
                                      //     - prec: outcome (measurement) precision, conditional on a fixed z-surface
                                      //     - beta: the regression coefficients (not currently supported--should be zero)
                                      //     - birth: add a new atom
                                      //     - move: change the location and height of one of the existing atoms
                                      //     - remove: remove one of the atoms
     double *dX,       double apre,  // dX is the step size for the proposal for each of the above move types.  apre and rpre
     double rpre,     double asigma,  // are the parameters of the Gamma prior distribution for the outcome (measurement) precision.
     double rsigma,                 // asigma,rsigma: the parameters of the gamma prior for the kernel precisions.
     double mbet,     double sbet,  // mbet,sbet: the parameters of the normal prior distribution for the intercept parameter.
     const Array1D<double> &pM,    // pM: the array of prior prob.s
				   // for values of M
     double agam,     double rgam,  // agam and rgam are parameters of the Levy process prior for the spatial 
                                      // surface.
     const Array1D<double> &temps,    // the temperatures for the
				      // parallel tempering process
     bool symm,                       // whether a symmetric gamma Levy random field is to be used
                                      // (otherwise a gamma Levy r.f. is used)
     bool printStats )                // Whether to print summary statistics about the MCMC
				      // (e.g. acceptance rates)
{
  int i;
  datX_ = Array1D<double>(datLen);
  datZ_ = Array1D<double>(datLen);
  for( i=0; i<datLen; i++ ){
    datX_[i] = datX[i];
    datZ_[i] = datZ[i];
  }

  #ifdef DEBUG1
    cout << "datX_ = \n" << datX_ << "\ndatZ_ = \n" << datZ_;
  #endif
  datLen_ = datLen;
  OutLG_ = OutLG;
  OutFL_ = OutFL;
  OutVL_ = OutVL;
  xlim_ = new double [2];
  xlim_[0]=xlim[0];
  xlim_[1]=xlim[1];
  ItnMax_ = ItnMax;
  ItnBrn_ = ItnBrn;
  SavMax_ = SavMax;
  pX_ = Array1D<double> (NmoveTypes);
  for( i=0; i<NmoveTypes; i++ )
    pX_[i] = pX[i];
  dX_= Array1D<double> (Nscales);
  for( i=0; i<Nscales; i++ )
    dX_[i] = dX[i];
  apre_ = apre;
  rpre_ = rpre;
  mbet_ = mbet;
  sbet_ = sbet;
  int pMlen = pM.dim();
  pM_ = Array1D<double> (pMlen);
  for( i=0; i<pMlen; i++ )
    pM_[i] = pM[i];
  agam_ = agam;
  rgam_ = rgam;
  asigma_ = asigma;
  rsigma_ = rsigma;
  symm_ = symm;
  temps_ = temps.copy();
  nTemps = temps_.dim();
  printStats_ = printStats;
  if( (agam_ <= 0) || (rgam_ <= 0) || (asigma_ <=0) || (rsigma_ <= 0) || (sbet_ <=0) ||
      (apre_ <= 0) || (rpre_ <= 0)){
    printf("Invalid hyperparameter specification\n");
    return;
  }

  // Calculate the constant $\log (\rho_\sigma^{\alpha_\sigma} / \Gamma(\alpha_\sigma) )$
  // needed for evaluation of the Levy density function $\nu(\gamma, s)$ as described in
  // the computation article.
  levyConst = asigma_ * log(rsigma_);
  levyConst = levyConst - lgamma(asigma_);  // lgamma is supposed to be the log-gamma function
}

LevyMCMC::~LevyMCMC(){}


/**************************************************************************
 **  Calculate the negative log likelihood 
 **  
 **/  
double LevyMCMC::nllh( LevyProc *levy,
		       double pre, double beta, double *SSE,
		       double temp ) {

  double startTime = tk.clockSecs();
  (*SSE) = levy->sumSquaredError( datX_, datZ_, beta );
  // We don't actually need to pass beta to the SSE function;
  // should just subtract beta from datZ_ before calling.

  double nlgLik = ((*SSE) * pre - datX_.dim() * log(pre))/2;
  
  likeEvalTime += tk.clockSecs() - startTime;

  return( nlgLik / temp );
}

/**************************************************************************
 **  Update the outcome precision parameter via a Gibbs full conditional 
 **  draw
 **/  
double LevyMCMC::updatePrecision( 
  LevyProc *levy,
  double beta,
  double *nlgLik, 
  double *sse,
  double temp ){

  double shape = apre_ + ((double) datX_.dim()) / (2*temp),
    rate = rpre_ + (*sse)/(2*temp);

  if( (rate <= 0) || (shape <= 0) ){
    MESSAGE "rate and shape must be positive\n" ERROR;
  }
  // generate from the full conditional dist'n
  // of the observation precision sigma^-2
  double pre  = S_rgamma( shape );
  pre = pre / rate;  
  
  // calc. the new negative log-likelihood and new SSE
  (*nlgLik) = nllh( levy, pre, beta, sse, temp );

  return(pre);                      // Always accept Gibbs step
}

/**************************************************************************
 **  Update the intercept parameter via a Gibbs full conditional 
 **  draw.
 **/  
double LevyMCMC::updateBeta( 
  LevyProc *levy, 
  double pre,
  double *nlgLik, 
  double *sse, 
  double temp ){

  double betaPre = datLen_ * pre / temp;
  betaPre = betaPre + 1/SQR(sbet_);

  double betaMean = levy->sumError( datX_, datZ_ );
  betaMean = betaMean * pre / temp;
  betaMean = betaMean + mbet_ / SQR(sbet_);
  betaMean = betaMean / betaPre;

  double betaSD = pow(betaPre, -0.5);
  double beta = S_rnorm();
  beta = beta * betaSD;
  beta = beta + betaMean;

  #ifdef DEBUG_BETA
    cout << "betaMean = " << betaMean << ", betaSD = " <<
      betaSD << ", new beta = " << beta << endl;
  #endif  
  // calc. the new negative log-likelihood and new SSE
  (*nlgLik) = nllh( levy, pre, beta, sse, temp );

  return(beta);                      // Always accept Gibbs step
}

/**************************************************************************
 **  Attempt to add a new atom to the Levy process.  If the addition is 
 **  accepted, update levy, nlgLik, and sse.
 **/  
bool LevyMCMC::proposeNewPoint
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse, double temp ){

  double x, sig2, z, pdf, nsse;
  levy->drawPoint( &x, &sig2, &z, dX_[SCA_DRAWPT], &pdf );    // New (proposed) point

  #ifdef DEBUG2
    printf("proposing x=%f, sig2=%f, z=%f, pdf = %f \n", x, sig2, z, pdf );
  #endif

  if( (sig2 <= 0) ){
    MESSAGE "Error: variance is non-positive.\n" ERROR;
  }
  double Padd = pX_[MV_NEWPT],
    Prem = pX_[MV_DELPT];

  int addInd = levy->addPoint( x, sig2, z );  // Add the new point to the Levy 
                            // process and keep its index for easy removal.

  // calculate the negative log likelihood and sum of squared errors of the 
  // proposal
  double newLik = nllh( levy, pre, beta, &nsse, temp );
  // calculate the difference in the log posterior probabilities minus
  // the difference in the proposal probabilities
  double logH = (*nlgLik) - newLik + log( Prem / Padd ) - 
    pdf + levy->dLevy(z, sig2);
  int Mnew = levy->length();
  if( pM_[Mnew] <= 0.0 )
    logH = -DBL_MAX;
  else   
    logH = logH + log( pM_[Mnew] ) - log( pM_[Mnew-1] );  

  #ifdef DEBUG2
    printAcceptProb(logH);
  #endif

  double samp = S_rexp();
  if( logH + samp > 0 ) {   // accept the proposed addition
    (*sse) = nsse;          // update the sum of sq. errors
    (*nlgLik) = newLik;     // update the neg. log like.
    #ifdef DEBUG2
      printf("proposal accepted\n");
    #endif
    return (true);
  } else {
    // Reject the proposed addition.  Remove the proposed point from the 
    // Levy process.
    levy->deletePoint( addInd );
    return (false);
  }
}


/**************************************************************************
 **  Attempt to change the location of an atom from the Levy
 **  process.  If the move is accepted, update levy, nlgLik, and
 **  sse.
 **/
bool LevyMCMC::moveLoc
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse,
  double temp ){
  
  int nPoints = levy->length();
  #ifdef DEBUG2
    printf("nPoints = %d\n", nPoints);
  #endif

  if (nPoints == 0)
    return(false);
  // Sample an atom for moving
  Array1D<double> equalProbs (nPoints, 1/(double)nPoints);
  #ifdef DEBUG3
    cout << "equalProbs:" << endl;
    cout << equalProbs << endl;
  #endif
  int j = rdiscrete(equalProbs);
  if( (j < 0) || (j >= nPoints) )
    MESSAGE "Choosing a nonexistent point to move\n" ERROR;

  #ifdef DEBUG2
    printf(" choosing atom %d for move\n", j);
  #endif

  // Get the location, variance, and height of the atom with index j
  double x, sig2, z;
  levy->getPoint( j, &x, &sig2, &z );

  // propose step in location
  double dx = S_rnorm() * dX_[SCA_MVLOC];
  levy->changePoint( j, (x+dx), sig2, z );
  
  // calculate the negative log likelihood and sum of squared errors of the 
  // proposal
  double nsse,
    newLik = nllh( levy, pre, beta, &nsse, temp );
  double logH = (*nlgLik) - newLik;
  #ifdef DEBUG2
    printAcceptProb(logH);
  #endif

  // Accept or reject
  double samp = S_rexp();
  if( logH + samp > 0 ) {   // accept the proposed addition
    (*sse) = nsse;          // update the sum of sq. errors
    (*nlgLik) = newLik;     // update the neg. log like.
    #ifdef DEBUG2
	printf("proposal accepted\n");
    #endif
    return (true);
  } else {
    // Reject the proposed addition.  Move the proposed point back to its
    // original location and height
    levy->changePoint( j, x, sig2, z );
    return (false);
  }
}

/**************************************************************************
 **  Attempt to change the magnitude of an atom from the Levy
 **  process.  If the move is accepted, update levy, nlgLik, and
 **  sse.
 **/
bool LevyMCMC::moveMag
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse, double temp ){
  
  int nPoints = levy->length();
  #ifdef DEBUG2
    printf("nPoints = %d\n", nPoints);
  #endif

  if (nPoints == 0)
    return(false);
  // Sample an atom for moving
  Array1D<double> equalProbs (nPoints, 1/(double)nPoints);
  #ifdef DEBUG3
    cout << "equalProbs:" << endl;
    cout << equalProbs << endl;
  #endif
  int j = rdiscrete(equalProbs);
  if( (j < 0) || (j >= nPoints) )
    MESSAGE "Choosing a nonexistent point to move\n" ERROR;

  #ifdef DEBUG2
    printf(" choosing atom %d for magnitude change\n", j);
  #endif

  // Get the location, variance, and height of the atom with index j
  double x, sig2, z;
  levy->getPoint( j, &x, &sig2, &z );

  double step = S_rnorm() * dX_[SCA_MVMAG],
    nz = z * exp(step);   // New z (atom magnitude)
  #ifdef DEBUG3
    cout << "  old z: " << z << 
      "  new z: " << nz << endl;
  #endif
  levy->changePoint( j, x, sig2, nz );

  // calculate the negative log likelihood and sum of squared errors of the 
  // proposal
  double nsse,
    newLik = nllh( levy, pre, beta, &nsse, temp );
  double logH = (*nlgLik) - newLik + agam_ * log( fabs(nz) / fabs(z) );
  logH = logH + rgam_ * (fabs(z) - fabs(nz));
  #ifdef DEBUG2
    printAcceptProb(logH);
  #endif

  // Accept or reject
  double samp = S_rexp();
  if( logH + samp > 0 ) {   // accept the proposed addition
    (*sse) = nsse;          // update the sum of sq. errors
    (*nlgLik) = newLik;     // update the neg. log like.
    #ifdef DEBUG2
	printf("proposal accepted\n");
    #endif
    return (true);
  } else {
    // Reject the proposed addition.  Move the proposed point back to its
    // original location and height
    levy->changePoint( j, x, sig2, z );
    return (false);
  }
}

/**************************************************************************
 **  Attempt to change the scale of an atom in the Levy process.  If
 **  the move is accepted, update levy, nlgLik, and sse.
 **/
bool LevyMCMC::changeScale
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse, double temp ){

  int nPoints = levy->length();
  #ifdef DEBUG2
    printf("nPoints = %d\n", nPoints);
  #endif

  if (nPoints == 0)
    return(false);
  // Sample an atom for a scale change
  Array1D<double> equalProbs (nPoints, 1/(double)nPoints);
  #ifdef DEBUG_SCALE
    cout << "equalProbs:" << endl;
    cout << equalProbs << endl;
  #endif
  int j = rdiscrete(equalProbs);
  if( (j < 0) || (j >= nPoints) )
    MESSAGE "Choosing a nonexistent point for scale change\n" ERROR;

  #ifdef DEBUG2
    printf(" choosing atom %d for scale change \n", j);
  #endif

  // Get the location, variance, and height of the atom with index j
  double x, sig2, z;
  levy->getPoint( j, &x, &sig2, &z );

  double step = S_rnorm() * dX_[SCA_CHSC],
    nSig2 = sig2 * exp(step);   // New sig2 (atom variance)  
  #ifdef DEBUG_SCALE
    cout << "step: " << step << "  old sig2: " << sig2 << 
      "  new sig2: " << nSig2 << endl;
  #endif

  // Try changing the scale of the atom
  levy->changePoint( j, x, nSig2, z );

  // calculate the negative log likelihood and sum of squared errors of the 
  // proposal
  double nsse,
    newLik = nllh( levy, pre, beta, &nsse, temp );
  double logH = (*nlgLik) - newLik + rsigma_ * (1/sig2 - 1/nSig2);
  logH = logH + asigma_ * log( sig2 / nSig2 );
  #ifdef DEBUG2
    printAcceptProb(logH);
  #endif

  // Accept or reject
  double samp = S_rexp();
  if( logH + samp > 0 ) {   // accept the proposed addition
    (*sse) = nsse;          // update the sum of sq. errors
    (*nlgLik) = newLik;     // update the neg. log like.
    #ifdef DEBUG2
	printf("proposal accepted\n");
    #endif
    return (true);
  } else {
    // Reject the proposed addition.  Move the proposed point back to its
    // original location and height
    levy->changePoint( j, x, sig2, z );
    return (false);
  }

}

/**************************************************************************
 **  Attempt to remove atom with indicator ind from the Levy process.  If 
 **  the removal is accepted, update levy, nlgLik, and sse.
 **/
bool LevyMCMC::deleteGivenPoint
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse, int ind, double temp ){

  // Delete the atom, saving the location and magnitude in case the 
  // proposed deletion is rejected.
  double x, sig2, z, nsse;
  levy->getPoint( ind, &x, &sig2, &z );
  #ifdef DEBUG2
    printf(" choosing atom (%f, %f, %f) for delete\n", x, sig2, z );
  #endif
  levy->deletePoint( ind );

  // calculate the negative log likelihood and sum of squared errors of the 
  // proposal
  double newLik = nllh( levy, pre, beta, &nsse, temp );

  // Calculate the density of the atom birth distribution, evaluated at
  // the deleted point
  double pdf = levy->dBirth( x, sig2, z, dX_[SCA_DRAWPT] );

  // Calculate the probability of proposing such a deletion and the 
  // probability of proposing the opposite addition.
  double Padd = pX_[MV_NEWPT],
    Prem = pX_[MV_DELPT];
  
  // calculate the difference in the log posterior probabilities minus
  // the difference in the proposal probabilities
  double logH = (*nlgLik) - newLik + log( Padd / Prem ) +
    pdf - levy->dLevy(z, sig2);
  int Mnew = levy->length();
  if( pM_[Mnew] <= 0.0 )
    logH = -DBL_MAX;
  else   
    logH = logH + log( pM_[Mnew] ) - log( pM_[Mnew+1] );  

  #ifdef DEBUG2
    printAcceptProb( logH );
  #endif 

  double samp = S_rexp();
  if( logH + samp > 0 ) {   // accept the proposed addition
    (*sse) = nsse;          // update the sum of sq. errors
    (*nlgLik) = newLik;     // update the neg. log like.
    #ifdef DEBUG2
      printf("proposal accepted\n");
    #endif
    return (true);
  } else {
    // Reject the proposed deletion.  Add the proposed point back into the 
    // Levy process.
    levy->addPoint( x, sig2, z, ind );
    return (false);
  }
}

/**************************************************************************
 **  Attempt to remove an atom from the Levy process.  If the removal is 
 **  accepted, update levy, nlgLik, and sse.
 **/
bool LevyMCMC::deletePoint
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse, double temp ){
  
  int nPoints = levy->length();
  #ifdef DEBUG2
    printf("nPoints = %d\n", nPoints);
  #endif

  if (nPoints == 0)
    return(false);
  // Sample an atom for deletion
  Array1D<double> equalProbs (nPoints, 1/(double)nPoints);
  int j = rdiscrete(equalProbs);       // sample the point to remove

  return( deleteGivenPoint( levy, pre, beta, nlgLik, sse, j, temp ) );
}


/**************************************************************************
 **  Attempt to split an atom in the Levy process.  If the split is 
 **  accepted, update levy, nlgLik, and sse.
 **/  
bool LevyMCMC::splitPoint
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse, double temp ){

  int nPoints = levy->length();
  if (nPoints == 0){
    #ifdef DEBUG2
      printf(" rejected split move: no atoms to split\n" );
    #endif
    return(false);
  }
  // Sample an atom for splitting
  Array1D<double> equalProbs (nPoints, 1/(double)nPoints);
  int ind = rdiscrete(equalProbs);       

  double x, sig2, z, x_1, sig2_1, z_1, x_2, sig2_2, z_2, pdf, nsse;
  levy->splitPoint( ind, dX_[SCA_SPL], &x, &sig2, &z, &x_1, &sig2_1, &z_1, &x_2, &sig2_2, &z_2, &pdf );

  #ifdef DEBUG2
    printf(" choosing atom (%f, %f, %f) for split\n", x, sig2, z );
  #endif
  levy->deletePoint( ind );  // Delete the old point from the process.
  int addInd1 = levy->addPoint( x_1, sig2_1, z_1 );  // Add the new points to the Levy 
  int addInd2 = levy->addPoint( x_2, sig2_2, z_2 );  // process and keep their indices for easy removal.  
  
  #ifdef DEBUG2
    printf("proposing x_1=%f, sig2_1=%f, z_1=%f, x_2=%f, sig2_2=%f, z_2=%f \n", x_1, sig2_1, z_1, x_2, sig2_2, z_2 );
  #endif

  if( (sig2_1 <= 0) || (sig2_2 <= 0) ){
    MESSAGE "Error: variance is non-positive.\n" ERROR;
  }

  double Pspl = pX_[MV_SPLIT],
    Pmer = pX_[MV_MERGE];

  // calculate the negative log likelihood and sum of squared errors of the 
  // proposal
  double newLik = nllh( levy, pre, beta, &nsse, temp );
  // calculate the difference in the log posterior probabilities minus
  // the difference in the log proposal probabilities
  double logH = (*nlgLik) - newLik + log( Pmer / Pspl ) - pdf;
  logH = logH + levy->dLevy( z_1, sig2_1 );
  logH = logH + levy->dLevy( z_2, sig2_2 );
  logH = logH - levy->dLevy( z, sig2 );
  int Mnew = levy->length();
  logH = logH + log( Mnew );
  if( pM_[Mnew] <= 0.0 )
    logH = -DBL_MAX;
  else   
    logH = logH + log( pM_[Mnew] ) - log( pM_[Mnew-1] );  

  #ifdef DEBUG2
    printAcceptProb(logH);
  #endif

  double samp = S_rexp();
  if( logH + samp > 0 ) {   // accept the proposed addition
    (*sse) = nsse;          // update the sum of sq. errors
    (*nlgLik) = newLik;     // update the neg. log like.
    #ifdef DEBUG2
      printf("proposal accepted\n");
    #endif
    return (true);
  } else {
    // Reject the proposed split.  Remove the proposed new points from the 
    // Levy process, and replace the removed point.  Do this in the 
    // reverse order from before, so that the add and delete indices are correct.
    levy->deletePoint( addInd2 );
    levy->deletePoint( addInd1 );
    levy->addPoint( x, sig2, z, ind );
    return (false);
  }
}


/**************************************************************************
 **  Attempt to merge an atom in the Levy process with the next atom to its
 **  right.  If the merge is accepted, update levy, nlgLik, and sse.
 **/
bool LevyMCMC::mergePoints
( LevyProc *levy, double pre, double beta,
  double *nlgLik,
  double *sse, double temp ){

  int nPoints = levy->length();
  if (nPoints <= 1){
    #ifdef DEBUG2
      printf(" rejected merge move: fewer than 2 atoms\n" );
    #endif
    return(false);
  }

  // Sample an atom for merging
  Array1D<double> equalProbs (nPoints, 1/(double)nPoints);
  int ind2, ind1 = rdiscrete(equalProbs);       

  double x, sig2, z, x_1, sig2_1, z_1, x_2, sig2_2, z_2, nsse;
  // merge the atom with index ind1 with the atom that is just to its
  // right.  If there is no atom to the right, resample ind1 until
  // there is such an atom.
  while( 0 > levy->mergePoints( ind1, &ind2, dX_[SCA_SPL], &x, &sig2, &z, &x_1, 
			    &sig2_1, &z_1, &x_2, &sig2_2, &z_2 ) ){
    // printf("Resampling merge index. ind1 = %d \n", ind1 );
    //  levy->print(); 
    ind1 = rdiscrete(equalProbs);       
  }

  #ifdef DEBUG2
    printf(" choosing atoms %d:(%f, %f, %f) and %d:(%f, %f, %f) for merge\n", 
	   ind1, x_1, sig2_1, z_1, ind2, x_2, sig2_2, z_2 );
  #endif
  // Delete the old points from the process.  Delete the one with the larger
  // index first, so that it does not change the index of the other.
  if( ind1 > ind2 ){
    levy->deletePoint( ind1 );  
    levy->deletePoint( ind2 );  
  } else {
    levy->deletePoint( ind2 );  
    levy->deletePoint( ind1 );  
  }
  int addInd = levy->addPoint( x, sig2, z );  // Add the new point 
  
  #ifdef DEBUG2
    printf("proposing x=%f, sig2=%f, z=%f, with index %d \n", x, sig2, z, addInd );
  #endif

  if( (sig2 <= 0) ){
    MESSAGE "Error: variance is non-positive.\n" ERROR;
  }

  double Pspl = pX_[MV_SPLIT],
    Pmer = pX_[MV_MERGE];

  // calculate the negative log likelihood and sum of squared errors of the 
  // proposal
  double newLik = nllh( levy, pre, beta, &nsse, temp );
  // calculate the difference in the log posterior probabilities minus
  // the difference in the log proposal probabilities
  double pdf = levy->dSplit(addInd, x, z_1, sig2_2, z_2, dX_[SCA_SPL]),
    logH = (*nlgLik) - newLik + log( Pspl / Pmer ) + pdf;
  logH = logH - levy->dLevy( z_1, sig2_1 );
  logH = logH - levy->dLevy( z_2, sig2_2 );
  logH = logH + levy->dLevy( z, sig2 );
  int Mnew = levy->length();
  logH = logH - log( Mnew+1 );
  if( pM_[Mnew] <= 0.0 )
    logH = -DBL_MAX;
  else   
    logH = logH + log( pM_[Mnew] ) - log( pM_[Mnew+1] );  

  #ifdef DEBUG2
    printAcceptProb(logH);
  #endif

  double samp = S_rexp();
  if( logH + samp > 0 ) {   // accept the proposed addition
    (*sse) = nsse;          // update the sum of sq. errors
    (*nlgLik) = newLik;     // update the neg. log like.
    #ifdef DEBUG2
      printf("proposal accepted\n");
    #endif
    return (true);
  } else {
    // Reject the proposed merge.  Remove the proposed new point from the 
    // Levy process, and replace the removed points.  Do this in the 
    // reverse order from before, so that the add and delete indices are correct.
    levy->deletePoint( addInd );
    if( ind1 > ind2 ){
      levy->addPoint( x_2, sig2_2, z_2, ind2 );
      levy->addPoint( x_1, sig2_1, z_1, ind1 );
    } else {
      levy->addPoint( x_1, sig2_1, z_1, ind1 );
      levy->addPoint( x_2, sig2_2, z_2, ind2 );
    }
    return (false);
  }
}


/**************************************************************************
 **  Run the MCMC
 **/  
void LevyMCMC::doMCMC(){

  if (SavMax_ <= 0) {
    printf("SavMax specified as non-positive, which is invalid.\n");
    return;
  }  

  //  Calculate the interval between state saves 
  int dSav = (int) floor( ItnMax_ / (double) SavMax_ );  
  if (dSav == 0)   dSav = 1;

  double startTime = tk.clockSecs();

  // Create arrays for the number of tries and acceptances for each
  // move type and temperature
  Ntry = Array2D<int> (nTemps, NmoveTypes, 0);  
  Nacc = Array2D<int> (nTemps, NmoveTypes, 0); 
  
  //  Initialize the parameters
  if ((rpre_ <= 0) || (apre_ <= 0)){
    printf( "prior hyperparameter rpre or apre has an invalid value: rpre = %f, apre = %f\n", rpre_, apre_ );
    return;
  }
  // create vectors of the outcome precision and intercept for each tempered chain
  Array1D<double> pre (nTemps, apre_ / rpre_),        
                        // Initialize outcome precision to prior mean
    beta (nTemps, mbet_);                      
                        // Initialize intercept to prior mean

  #ifdef DEBUG1
    printf( "ItnBrn = %d, ItnMax = %d, dSav = %d, datLen_ = %d\n", 
	    ItnBrn_, ItnMax_, dSav, datLen_ );
    cout << "Pre: " << pre << endl << "beta: " << beta << endl;
    printf("OutLG = %s, OutFL = %s, OutVL = %s\n", OutLG_, OutFL_, OutVL_);
  #endif

  // Create a vector of Levy processes, one for each tempered chain.
  LevyProc *levy [nTemps];
  int i;
  for( i=0; i<nTemps; i++ ){
    // for each temperature
    // initialize the levy process by drawing from its prior dist'n
    levy[i] = new LevyProc(xlim_, pM_, agam_, rgam_, asigma_, rsigma_, symm_); 
    levy[i]->drawFromPrior();   
  }
  
  #ifdef DEBUG1
    printf("Initial Levy processes = \n");
    for( i=0; i<nTemps; i++ ){
      cout << i << endl;
      levy[i]->print();  
    }
    fflush(stdout);
  #endif
  
  // Create vectors to hold the negative log-likelihood and SSE for 
  // each tempered chain.
  Array1D<double> nlgLik (nTemps),
    sse (nTemps);
  for( i=0; i<nTemps; i++ ){
    // for each temperature
    // Force accep of 1st proposal
    nlgLik[i] = DBL_MAX;                           
    // Calculate the initial sum of squared errors
    if (datLen_ == 0)
      sse[i] = 0;
    else
      sse[i] = sum((datZ_-beta[i])*(datZ_-beta[i]));   
  }
  
  #ifdef DEBUG1
    cout <<"initial sse = \n" << sse << endl;
  #endif
  
  int mv, tempMv, swap;
  bool acc;
  
  FILE *LGfile = fopen(OutLG_, "w" );
  FILE *FLfile = fopen(OutFL_, "w" );
  FILE *VLfile = fopen(OutVL_, "w" );

  #ifdef DEBUG1
    printf("pX_ = %f %f %f %f %f %f %f %f\n", pX_[MV_PRE], pX_[MV_BETA], 
	   pX_[MV_NEWPT], pX_[MV_MVLOC], pX_[MV_MVMAG], pX_[MV_CHSC], 
           pX_[MV_DELPT], pX_[MV_SPLIT], pX_[MV_MERGE]);
    printf("dX_ = %f %f %f %f %f\n", dX_[SCA_DRAWPT],
	   dX_[SCA_MVMAG], dX_[SCA_CHSC], dX_[SCA_MVLOC], dX_[SCA_SPL]);
  #endif
  
  int itn = (-ItnBrn_-1);
  double nAtoms, itnDbl, samp, tempDbl, accProb;
  LevyProc *tempLevy;

  likeEvalTime = 0;   // Set the time spent on likelihood evaluations to zero

  // Create a vector specifying the probability of updating each tempered chain  
  Array1D<double> tempProb( nTemps, 1/(double)nTemps );
  // Create a vector swapping each tempered chain with the next-highest temp.
  Array1D<double> swapProb( (nTemps-1), 1/(double)(nTemps-1) );

  Array1D<int> nTrySwap( nTemps-1, 0 ),
    nAccSwap( nTemps-1, 0 );

  int *thisTry, *thisAcc;
  while( itn < ItnMax_ ){
    
    //  Generate a proposal:
    //  Pick a parameter to update
    mv = rdiscrete(pX_);     // Which type?
    //  Pick a tempered chain to update
    tempMv = rdiscrete( tempProb );
    Ntry[tempMv][mv] += 1;  // Increment the number of tries for this 
                            // temperature and move type

    switch (mv){
    case MV_PRE:
      pre[tempMv] = updatePrecision( levy[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
				     &(sse[tempMv]), temps_[tempMv] );
      #ifdef DEBUG2
        printf("updating precision for temp %f to %f\n", temps_[tempMv], pre[tempMv]);
      #endif
      Nacc[tempMv][MV_PRE] += 1;   // All precision moves are accepted
      break;
    case MV_BETA:
      beta[tempMv] = updateBeta( levy[tempMv], pre[tempMv], &(nlgLik[tempMv]), 
				 &(sse[tempMv]), temps_[tempMv] );
      #ifdef DEBUG2
        printf("updating beta for temp %f to %f\n", temps_[tempMv], beta[tempMv]);
      #endif
      Nacc[tempMv][MV_BETA] += 1;   // All beta moves are accepted
      break;
    case MV_NEWPT:
      acc = proposeNewPoint( levy[tempMv], pre[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
			     &(sse[tempMv]), temps_[tempMv] );
      if (acc) Nacc[tempMv][MV_NEWPT] += 1;
      break;
    case MV_MVLOC:
      acc = moveLoc( levy[tempMv], pre[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
		     &(sse[tempMv]), temps_[tempMv] );
      if (acc) Nacc[tempMv][MV_MVLOC] += 1;
      break;
    case MV_MVMAG:
      acc = moveMag( levy[tempMv], pre[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
		     &(sse[tempMv]), temps_[tempMv] );
      if (acc) Nacc[tempMv][MV_MVMAG] += 1;
      break;
    case MV_CHSC:
      acc = changeScale( levy[tempMv], pre[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
			 &(sse[tempMv]), temps_[tempMv] );
      if (acc) Nacc[tempMv][MV_CHSC] += 1;
      break;
    case MV_DELPT:
      acc = deletePoint( levy[tempMv], pre[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
			 &(sse[tempMv]), temps_[tempMv] );
      if (acc) Nacc[tempMv][MV_DELPT] += 1;
      break;
    case MV_SPLIT:
      acc = splitPoint( levy[tempMv], pre[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
			&(sse[tempMv]), temps_[tempMv] );
      if (acc) Nacc[tempMv][MV_SPLIT] += 1;
      break;
    case MV_MERGE:
      acc = mergePoints( levy[tempMv], pre[tempMv], beta[tempMv], &(nlgLik[tempMv]), 
			 &(sse[tempMv]), temps_[tempMv] );
      if (acc) Nacc[tempMv][MV_MERGE] += 1;
      break;
    default:
      printf("lrfRun: invalid move type\n");
    }
    #ifdef DEBUG1
      cout << "Pre: " << pre << endl << "beta: " << beta << endl;
      cout << "Sse: " << sse << endl << "nlgLik: " << nlgLik << endl;
      printf( "levy process for temp. %f = \n", temps_[tempMv] );
      levy[tempMv]->print();  fflush(stdout);
    #endif

    // attempt a swap 
    if((nTemps > 1) && (S_runif() < 0.1)){
      //  Pick a tempered chain to swap with the next-highest temp.
      swap = rdiscrete( swapProb );
      #ifdef DEBUG_SWAP
        cout << "Attempting to swap chain " << swap << " with next-highest temp." << endl;
	cout << "Chain " << swap << " parameters:  beta= " << beta[swap] << "  pre= " << 
	  pre[swap] << " Levy= " << endl;
	levy[swap]->print();
	cout << "Chain " << (swap+1) << " parameters:  beta= " << beta[swap+1] << "  pre= " << 
	  pre[swap+1] << " Levy= " << endl;
	levy[swap+1]->print();
      #endif
      nTrySwap[swap]++;
      // calculate the acceptance prob. of the swap
      accProb = nllh( levy[swap], pre[swap], beta[swap], &tempDbl, temps_[swap] );
      accProb += nllh( levy[swap+1], pre[swap+1], beta[swap+1], &tempDbl, temps_[swap+1] );
      accProb -= nllh( levy[swap], pre[swap], beta[swap], &tempDbl, temps_[swap+1] );
      accProb -= nllh( levy[swap+1], pre[swap+1], beta[swap+1], &tempDbl, temps_[swap] );
  
      samp = S_rexp();
      if( accProb + samp > 0 ) {   
	nAccSwap[swap]++;
        // with probability exp(accProb)
        // accept the proposed swap
        tempLevy = levy[swap];
        levy[swap] = levy[swap+1];
        levy[swap+1] = tempLevy;
        tempDbl = beta[swap];
        beta[swap] = beta[swap+1];
        beta[swap+1] = tempDbl;
        tempDbl = pre[swap];
        pre[swap] = pre[swap+1];
        pre[swap+1] = tempDbl;
        tempDbl = sse[swap];
        sse[swap] = sse[swap+1];
        sse[swap+1] = tempDbl;
        // update the neg. log like.
        nlgLik[swap] = nllh( levy[swap], pre[swap], beta[swap], &tempDbl, temps_[swap] );
        nlgLik[swap+1] = nllh( levy[swap+1], pre[swap+1], beta[swap+1], &tempDbl, temps_[swap+1] );
  
        #ifdef DEBUG_SWAP
          printf("swap accepted\n");
	cout << "Chain " << swap << " new parameters:  beta= " << beta[swap] << "  pre= " << 
	  pre[swap] << " Levy= " << endl;
	levy[swap]->print();
	cout << "Chain " << (swap+1) << " new parameters:  beta= " << beta[swap+1] << "  pre= " << 
	  pre[swap+1] << " Levy= " << endl;
	levy[swap+1]->print();
        #endif
      } 
    }
    if( ((itn % dSav) == 0) && (itn >= 0) ){
      // Write parameters to file
      itnDbl = (double) itn;
      fwrite( &itnDbl, sizeof(double), 1, FLfile );

      for( i=0; i<nTemps; i++ ){
	// for each temperature
        nAtoms = (double) levy[i]->length();
        fwrite( &nAtoms, sizeof(double), 1, FLfile );
        fwrite( &(sse[i]), sizeof(double), 1, FLfile );
        fwrite( &(pre[i]), sizeof(double), 1, FLfile );
	fwrite( &(beta[i]), sizeof(double), 1, FLfile );
        levy[i]->appendToFile( VLfile );
      }
    }
    itn++;
  }



  // --------------------------   Wrap Up  -----------------------------
  fclose( FLfile );
  fclose( VLfile );
  if(itn>0){
    double compTime = tk.clockSecs() - startTime;
    if(printStats_){
      //    printf( "Completed %d burn-in and %d steps in %f seconds\n", ItnBrn_, itn,
      //       currentTime-Start );
      printf( "Total computation time: %f\n", compTime );
      printf( "Computation time for likelihood evals: %f\n", likeEvalTime );
      cout << "Number of Tries: " << Ntry << endl << "Number of Accepts: " << Nacc << endl;
      cout << "Number of swaps attempted: " << nTrySwap << endl << "  and accepted: " << nAccSwap << endl;
    }
    // Write the same information to a log file.  Write it in R format so that it can be read into R/S-PLUS using
    // the dget command
    fprintf( LGfile, "list(\"compTime\" = %f\n", compTime );
    fprintf( LGfile, ", \"compTimeLike\" = %f\n", likeEvalTime );
    for( i=0; i<nTemps; i++ ){
      thisTry = Ntry[i];
      thisAcc = Nacc[i];
      fprintf( LGfile, ", \"numTriesTemp%d\"  = c(%d, %d, %d, %d, %d, %d, %d, %d, %d)\n", i, thisTry[MV_PRE], thisTry[MV_BETA],
	       thisTry[MV_NEWPT], thisTry[MV_MVLOC], thisTry[MV_MVMAG], thisTry[MV_CHSC], thisTry[MV_DELPT], thisTry[MV_SPLIT], 
	       thisTry[MV_MERGE]);
      fprintf( LGfile, ", \"numAcceptsTemp%d\" = c(%d, %d, %d, %d, %d, %d, %d, %d, %d)\n)\n", i, thisAcc[MV_PRE], thisAcc[MV_BETA],
	       thisAcc[MV_NEWPT], thisAcc[MV_MVLOC], thisAcc[MV_MVMAG], thisAcc[MV_CHSC], thisAcc[MV_DELPT], thisAcc[MV_SPLIT], 
	       thisAcc[MV_MERGE]);
    }
  }
  fclose( LGfile );
  
}


