#ifndef LEVY_MCMC_H
#define LEVY_MCMC_H


#include "tnt_array2d.h"
#include "tnt_array1d.h"
#include "tnt_array1d_utils.h"
#include "tnt_array2d_utils.h"
#include "moreArray1dUtils.h"
#include "randGen.h"
#include "LevyProc.h"


using TNT::Array1D;
using TNT::Array2D;


class LevyMCMC{
 private:
  Array1D<double> datX_, datZ_;
  int datLen_;
  char *OutLG_, *OutFL_, *OutVL_;
  double *xlim_;
  int ItnMax_, ItnBrn_, 
    SavMax_;
  Array1D<double> pX_, dX_;
  Array1D<double> pM_;
  double apre_, rpre_, asigma_, rsigma_, mbet_, sbet_, 
    agam_, rgam_, levyConst;
  bool symm_;  // whether a symmetric gamma Levy random field is to be used
              // (otherwise a gamma Levy r.f. is used)

  Array1D<double> temps_;           // the temperatures for the
  int nTemps;       		    // parallel tempering process

  double likeEvalTime;
  timeKeeper tk;

  Array2D<int> Ntry,  // The number of tries and
    Nacc;                //  acceptances of various move types
  bool printStats_;     // Whether to print MCMC summaries
			// such as acceptance rates.
  
  /**************************************************************************
   **  Calculate the negative log likelihood and update the sum of
   **  squared errors.
   **/  
  double nllh( LevyProc *levy, double pre, double beta, double *SSE, 
	       double temp );
  
  /**************************************************************************
   **  Update the outcome precision parameter via a Gibbs full conditional 
   **  draw
   **/  
  double updatePrecision( LevyProc *levy, double beta, 
			  double *nlgLik, double *sse, double temp );

  /**************************************************************************
   **  Update the intercept parameter beta via a Gibbs full conditional 
   **  draw
   **/  
  double updateBeta( LevyProc *levy, double pre,
		     double *nlgLik, double *sse, double temp );
  
  /**************************************************************************
   **  Attempt to add a new atom to the Levy process.  If the addition is 
   **  accepted, update levy, nlgLik, and sse.
   **/  
  bool proposeNewPoint( LevyProc *levy, double pre, double beta, 
                        double *nlgLik, double *sse, double temp );
  
  /**************************************************************************
   **  Attempt to change the location of an atom from the Levy
   **  process.  If the move is accepted, update levy, nlgLik, and
   **  sse.
   **/
  bool moveLoc( LevyProc *levy, double pre, double beta, 
		  double *nlgLik, double *sse, double temp );

  /**************************************************************************
   **  Attempt to change the magnitude of an atom from the Levy
   **  process.  If the move is accepted, update levy, nlgLik, and
   **  sse.
   **/
  bool moveMag( LevyProc *levy, double pre, double beta, 
		  double *nlgLik, double *sse, double temp );

  /**************************************************************************
   **  Attempt to change the scale of an atom from the Levy process.
   **  If the move is accepted, update levy, nlgLik, and sse.
   **/
  bool changeScale( LevyProc *levy, double pre, double beta, 
		  double *nlgLik, double *sse, double temp );

  /**************************************************************************
   **  Attempt to remove an atom from the Levy process.  If the removal is 
   **  accepted, update levy, nlgLik, and sse.
   **/
  bool deletePoint( LevyProc *levy, double pre, double beta, 
		    double *nlgLik, double *sse, double temp );

  /**************************************************************************
   **  Attempt to remove the atom with indicator ind from the Levy
   **  process.  If the removal is accepted, update levy, nlgLik, and
   **  sse.
   **/
  bool deleteGivenPoint( LevyProc *levy, double pre, double beta, 
			 double *nlgLik, double *sse, int ind, double temp );

  /**************************************************************************
   **  Attempt to split an atom in the Levy process.  If the split is 
   **  accepted, update levy, nlgLik, and sse.
   **/  
  bool splitPoint( LevyProc *levy, double pre, double beta, 
                        double *nlgLik, double *sse, double temp );

  /**************************************************************************
   **  Attempt to merge adjacent atoms in the Levy process.  If the merge is 
   **  accepted, update levy, nlgLik, and sse.
   **/  
  bool mergePoints( LevyProc *levy, double pre, double beta, 
                        double *nlgLik, double *sse, double temp );
  
 public:
  LevyMCMC(
     double *datX,                    // The data (x, z for each of datLen samples, where x is the location
     double *datZ,   int datLen,   // and z is the measurement)
     char *OutLG,                     // The name of the log file
     char *OutFL,    char *OutVL,     // The names of the fixed-length and variable-length files to be written
     double *xlim,                    // The bounding region within which to fit the surface.  xlim should 
                                      //be length-2 vectors
     int ItnMax,     int ItnBrn,  // ItnMax is the number of total iterations after burnin, while ItnBrn is 
                                      // the length of the burnin
     int SavMax,     double *pX,    // SavMax is the maximum number of saved iterations.  The resulting thinning 
                                      // is approximately 
                                      // ItnMax / SavMax.  pX is a vector of the probabilities of proposing a move 
                                      // for each of the parameters:
                                      //     - prec: outcome (measurement) precision, conditional on a fixed z-surface
                                      //     - beta: the regression coefficients (not currently supported--should be zero)
                                      //     - birth: add a new atom
                                      //     - move: change the location and height of one of the existing atoms
                                      //     - remove: remove one of the atoms
     double *dX,     double apre,  // dX is the step size for the proposal for each of the above move types.  apre and rpre
     double rpre,    double asigma,  // are the parameters of the Gamma
     double rsigma,		   // prior distribution for the
				   // outcome (measurement) precision.
				   // asigma,rsigma: the parameters of
				   // the Gamma prior for the kernel
				   // precision parameters.
     double mbet,    double sbet,  // mbet,sbet: the parameters of the normal prior distribution for the intercept parameter.
     const Array1D<double> &pM,    // pM: the array of prior prob.s
				   // for values of M
     double agam,    double rgam,  // agam and rgam are the parameters of the Levy process prior for the spatial 
                                      // surface.
     const Array1D<double> &temps,    // the temperatures for the
				      // parallel tempering process
     bool symm = true,                // whether a symmetric gamma Levy random field is to be used
                                      // (otherwise a gamma Levy r.f. is used)
     bool printStats = true);         // Whether to print summary
				      // statistics about the MCMC
				      // (e.g. acceptance rates)

    ~LevyMCMC();

  /**************************************************************************
   **  Run the MCMC
   **/  
  void doMCMC();

};

#endif
