
#include "LevyProc.h"
#include "randGen.h"

#define SQR(x) ((x)*(x))
#define SQRT(x) pow(x,0.5)

extern "C"
{
  double S_STDCALL S_rpois( double lambda );
  double S_STDCALL S_runif( void );
  double S_STDCALL S_rexp( void );
}

using std::cout;
using std::endl;

/**************************************************************************
 **  e1 and e1inv
 **  
 **  Exponential Integral Function and its inverse (see A+S sec 5.1) 
 **  Note this is only used to draw the initial MCMC draw, so it
 **  is not critical.... the qchisq/pchisq approx'ns are good enough.
 **  The following work well for small x, large y... say, for x < 1 < y.
 **
 **/
double e1( double x ) {      // from A&S 5.1.53:
  return( -log(x) - 0.57721566 + x*(0.99999193 - x*(0.24991055 - 
    x*(0.55199680 - x*(0.00976004 - x*(0.00107857))))) );
}
double e1inv( double y ) {   // AIJM van Dijk (2002)  
  double t1 = exp(-y-0.577215665),
    t2 = exp(-2.216*y-0.722),
    t3 = exp(-6.547*y-2.481);
  return(t1+t2-t3);
}



/**************************************************************************
 **  Constructor
 **
 **/
LevyProc::LevyProc(
  double *xlim, 
  const Array1D<double> &pM, double agam, 
  double rgam, double asigma,
  double rsigma, bool symm
  )
{
  agam_ = agam;
  rgam_ = rgam;
  asigma_ = asigma;
  rsigma_ = rsigma;
  symm_ = symm;
  pM_ = pM.copy();

  xlim_ = new double [2];
  if( xlim == NULL ){         
    // If region not specified, default to unit square.
    xlim_[0] = 0;
    xlim_[1] = 1;
  } else {
    xlim_[0] = xlim[0];
    xlim_[1] = xlim[1];
  }
  area = ( xlim_[1] - xlim_[0] );
  #ifdef DEBUG1  
    printf("area = %f\n", area);
  #endif
  atomHeight_ = vector<double> ();
  atomVar_ = vector<double> ();
  xLoc_ = vector<double> ();

  nAtoms_ = 0;

  // Calculate the constant $\log (\rho_\sigma^{\alpha_\sigma} / \Gamma(\alpha_\sigma) )$
  // needed for evaluation of the birth density function $b(\gamma, s)$
  sigmaConst = asigma_ * log(rsigma_);
  sigmaConst = sigmaConst - lgamma(asigma_);
  // Calculate the constant $\log (\rho^{\alpha} / \Gamma(\alpha) )$
  // needed for evaluation of the birth density function $b(\gamma, s)$
  gamConst = agam_ * log(rgam_);
  gamConst = gamConst - lgamma(agam_);
}

/**************************************************************************
 **  Destructor
 **
 **/
LevyProc::~LevyProc(){}

/**************************************************************************
 **  Draw a location uniformly from the rectangular region.  Put the coords 
 **  in x.
 **
 **/
void LevyProc::drawUnifLoc(double *x){

  (*x) = xlim_[0] + (xlim_[1] - xlim_[0]) * S_runif();
}

/**************************************************************************
 **  Draw from the Gamma random field prior
 **/
void LevyProc::drawFromPrior() {
  
  #ifdef DEBUG1  
    printf( "agam = %f; rgam = %f; area = %f, symm = %d\n", agam_, rgam_, area, symm_);
  #endif
  int nAtomsToAdd = rdiscrete( pM_ );
  nAtomsToAdd = std::max( nAtomsToAdd, 1 );

  double x, sig2, z, pdf;
  for (int i=0; i<nAtomsToAdd; i++){
    
    drawPoint( &x, &sig2, &z, 1.0, &pdf );    
    addPoint( x, sig2, z ); 
  }
  #ifdef DEBUG1
    printf( "Natoms = %d\n", nAtoms_);
  #endif
}


/**************************************************************************
 **  Calculate the squared distance of z from the height of the Levy
 **  process surface at the location x, on the original or log scale.
 **
 **/
double LevyProc::sumSquaredError
( const Array1D<double> &x, 
  const Array1D<double> &z, 
  double beta,
  bool logScale ){

  int i, nDat = x.dim();
  double sse=0;
  double d2, YHat;

  vector<double>::iterator itX,
    itHeight, itSig2;

  for(i=0; i<nDat; i++){
    itSig2 = atomVar_.begin();  // first element of the vector
    itHeight = atomHeight_.begin();  // first element of the vector
    YHat = beta;
    
    // The slow part of the following calculation may be the 
    // exponential function
    for( itX = xLoc_.begin(); itX < xLoc_.end(); itX++ ){
      
      d2 = SQR((*itX) - x[i]);
      if( (*itSig2) <= 0.0 ){
	MESSAGE "kernel variance <= 0\n" ERROR;
      }

      #ifdef ALT_KERN
        YHat += (*itHeight) * pow(1 + SQRT(d2) / SQRT(*itSig2), -4);
      #else
        YHat += (*itHeight) * exp( -0.5 * d2 / (*itSig2) );
      #endif
      itHeight++;
      itSig2++;
    }
    if( logScale ){
      if( ( z[i] <= 0.0 ) || ( YHat <= 0.0 ) ){
        cout << "z[i] = " << z[i] << "  YHat = " << YHat << endl;  
        cout << "atomHeight = ";
        for( itHeight = atomHeight_.begin(); itHeight < atomHeight_.end(); itHeight++ )
  	cout << " " << (*itHeight);
        cout << endl;
        fflush(stdout);
        MESSAGE "z[i]<=0 or Yhat <= 0 in log-scale SSE calculation\n" ERROR;
      }  
      sse += SQR( log( z[i] / YHat ) );
    }
    else
      sse += SQR( z[i] - YHat );
  }
  return(sse);
}


/**************************************************************************
 **  Calculate the squared distance of z from the height of the Levy
 **  process surface at the location x, on the original or log scale.
 **
 **/
double LevyProc::sumError
( const Array1D<double> &x, 
  const Array1D<double> &z, 
  double beta,
  bool logScale ){

  int i, nDat = x.dim();
  double se=0;
  double d2, YHat;

  vector<double>::iterator itX,
    itHeight, itSig2;

  for(i=0; i<nDat; i++){
    itSig2 = atomVar_.begin();
    itHeight = atomHeight_.begin();
    YHat = beta;
    
    // The slow part of the following calculation may be the 
    // exponential function
    for( itX = xLoc_.begin(); itX < xLoc_.end(); itX++ ){
      
      d2 = SQR((*itX) - x[i]);
      if( (*itSig2) <= 0.0 ){
	MESSAGE "kernel variance <= 0\n" ERROR;
      }
      YHat += (*itHeight) * exp( -0.5 * d2 / (*itSig2) );
      itSig2++;
      itHeight++;
    }
    if( logScale ){
      if( ( z[i] <= 0.0 ) || ( YHat <= 0.0 ) ){
        cout << "z[i] = " << z[i] << "  YHat = " << YHat << endl;  
        cout << "atomHeight = ";
        for( itHeight = atomHeight_.begin(); itHeight < atomHeight_.end(); itHeight++ )
  	cout << " " << (*itHeight);
        cout << endl;
        fflush(stdout);
        MESSAGE "z[i]<=0 or Yhat <= 0 in log-scale error calculation\n" ERROR;
      }  
      se += ( log( z[i] / YHat ) );
    }
    else
      se += (z[i] - YHat);
  }
  return(se);
}

/**************************************************************************
 **  Draw a new atom for the Levy process.  Draw the location x
 **  uniformly from the rectangular region, and draw the height z
 **  from its prior.  Calculate the pdf (if ln = false) or the
 **  log-pdf (if ln = true).  Do not add the atom to the levy process.
 **/  
void LevyProc::drawPoint
( double *x, double *sig2, double *z,  // The location, variance, and
				       // height of the new atom
  double s,       // the proposal scale factor (the proposal is drawn
                  // from its prior, then multiplied by this scale.)
  double *pdf,     // The pdf
  bool ln        // Whether to return the log pdf or the pdf
  ) {

  drawUnifLoc( x );
  (*z) = S_rgamma( agam_ );     // New mass  z ~ Ga( agam, rgam / s )
  (*z) = (*z) * s / rgam_; 
  if( symm_ && (S_runif() < 0.5) )
    (*z) = -1 * (*z);

  (*sig2) = S_rgamma( asigma_ );   // sample the kernel variance
  (*sig2) = (*sig2) / rsigma_;        // from its inverse gamma prior.
  (*sig2) = 1 / (*sig2);
  
  (*pdf) = dBirth( (*x), (*sig2), (*z), s, ln);
}

/**************************************************************************
 ** Calculate the density of the atom birth proposal distribution
 **/  
double LevyProc::dBirth 
( double x, double sig2, double z,  // The location, variance, and
				    // height of the atom
  double s,      // the proposal scale factor (the proposal is drawn
                  // from its prior, then multiplied by this scale.)
  bool ln        // Whether to return the log pdf or the pdf
  ) {

  // First calc. the part of the birth density for the location param.
  double pdf = - log( area );
  // Now the part for the magnitude param.
  pdf = pdf + gamConst;
  pdf = pdf - agam_ * log(s);
  pdf = pdf + (agam_ - 1) * log( fabs(z) );
  pdf = pdf - rgam_ * fabs(z) / s;
  if(symm_) pdf = pdf - log( 2 );
  // Now the part of the birth density for the scale param.
  pdf = pdf + sigmaConst;
  pdf = pdf - (asigma_ + 1) * log( sig2 );
  pdf = pdf - rsigma_ / sig2;

  if(!ln) (pdf) = exp(pdf);
  return( pdf );
}

/**************************************************************************
 **  Evaluate the Levy density \nu/\nu_+ at (z, mu, sigma^2).  This quantity
 **  does not depend on mu.
 **/  
double LevyProc::dLevy 
( double z, double sig2,
  bool ln) {
  // Levy density at x,sig2,z (or LOG density if ln=TRUE) 
  // First the uniform density for mu
  double lpdf = - log( area );
 
  // Now the part for the magnitude param.
  lpdf = lpdf + gamConst;
  lpdf = lpdf + (agam_ - 1) * log( fabs(z) );
  lpdf = lpdf - rgam_ * fabs(z);
  if(symm_) lpdf = lpdf - log( 2 );
  // Now the part of the Levy density for the scale param.
  lpdf = lpdf + sigmaConst;
  lpdf = lpdf - (asigma_ + 1) * log( sig2 );
  lpdf = lpdf - rsigma_ / sig2;

  if (ln) return( lpdf );
  else return( exp(lpdf) );
}

/**************************************************************************
 **  Add an atom at location x and with variance sig2 and height z to
 **  the Levy process so that the atom has index ind in the new list
 **  of atoms.  ind should be between -1 and length() inclusive; a
 **  value of -1 is taken to be equal to length().  Return the index
 **  of the new atom.
 **/  
int LevyProc::addPoint
( double x, double sig2, double z,  // The location, variance, and height 
  int ind                           // of the new atom
  ) {

  if( (ind < -1) || (ind > nAtoms_) ){
    MESSAGE "Invalid atom index for insertion\n" ERROR;
  }

  if( (ind < nAtoms_) && (ind >= 0) ){
    vector<double>::iterator iter = xLoc_.begin() + ind;
    xLoc_.insert( iter, x );  
    iter = atomVar_.begin() + ind;
    atomVar_.insert( iter, sig2 );
    iter = atomHeight_.begin() + ind;
    atomHeight_.insert( iter, z );
  } else {
    // vector insert does not work for inserting at the end of the vector;
    // use push_back for this special case.
    xLoc_.push_back( x );
    atomVar_.push_back( sig2 );
    atomHeight_.push_back( z );
  }
  nAtoms_ += 1;

  if (ind == -1)
    return( nAtoms_-1 );
  else
    return( ind );
}

/**************************************************************************
 **  Get the location, variance, and magnitude of the atom with index ind.
 **/  
void LevyProc::getPoint
( int ind, double *x, double *sig2, double *z
  ) {
  if( (ind < 0) || (ind >= nAtoms_) ){
    MESSAGE "Invalid atom indicator\n" ERROR;
  }
  (*x) = xLoc_[ind];
  (*sig2) = atomVar_[ind];
  (*z) = atomHeight_[ind];
}


/**************************************************************************
 **  Change the atom with index ind to location x, variance sig2, and
 **  height z
 **/  
void LevyProc::changePoint
( int ind, double x, double sig2, double z
  ) {

  if( (ind < 0) || (ind >= nAtoms_) )
    MESSAGE "Invalid atom indicator\n" ERROR;
  if( sig2 <= 0 )
    MESSAGE "Invalid atom variance (<=0)\n" ERROR;

  // If necessary, reflect the atom location to be inside the limits
  while( (x < xlim_[0]) || (x > xlim_[1]) ) {
    if( x < xlim_[0] ) 
      x = 2 * xlim_[0] - x;   
    if( x > xlim_[1] ) 
      x = 2 * xlim_[1] - x;
  }

  // change the atom location and magnitude
  xLoc_[ind] = x;
  atomVar_[ind] = sig2;
  atomHeight_[ind] = z;
}

/**************************************************************************
 **  Delete the atom with index ind.
 **/  
void LevyProc::deletePoint( 
  int ind
  ) {
  if( (ind < 0) || (ind >= nAtoms_) ){
    MESSAGE "Invalid atom indicator for deletion\n" ERROR;
  }
  vector<double>::iterator iter = xLoc_.begin() + ind;
  xLoc_.erase( iter );
  iter = atomVar_.begin() + ind;
  atomVar_.erase( iter );
  iter = atomHeight_.begin() + ind;
  atomHeight_.erase( iter );

  nAtoms_ -= 1;
}


/**************************************************************************
**  Obtain the distance between the location of a particular atom
**  and the smaller of: the location of the next atom to the right,
**  or the righthand boundary of the region.  Used in proposing an
**  atom split.  x is the location of the atom of interest.
**/
double LevyProc::distAboveAtom(int ind){

  double x, sig2, z;
  getPoint( ind, &x, &sig2, &z );

  // find the smallest positive location difference
  double diff, minPosDiff = DBL_MAX;
  for( int i=0; i<nAtoms_; i++ ){
    diff = xLoc_[i] - x;
    if( (xLoc_[i] >= x) && (diff < minPosDiff) && (i != ind) )
      minPosDiff = diff;
  }

  double borderDiff = xlim_[1] - x,
    delta_mu = std::min( minPosDiff, borderDiff );

  if( delta_mu == 0 ){
    printf("Two atoms at the same location, or an atom on the righthand boundary.  Levy process:\n");
    this->print();
    MESSAGE "" ERROR;
  }
 
  return(delta_mu);
}

/**************************************************************************
 ** Calculate the density of the atom split proposal distribution
 **/  
double LevyProc::dSplit 
(int ind,  // the index of the split atom 
 double x, // The location of the split atom
 double z_1,  // The height of the first new atom
 double sig2_2, double z_2,  // The variance and
                             // height of the second new atom
 double s,      // the proposal scale for the magnitude step (delta_gamma)
 bool ln        // Whether to return the log pdf or the pdf
 ) {

  double delta_mu = distAboveAtom(ind),
    pdf = - log( delta_mu );

  // Now the density of the normal proposal for the magnitude step
  double zDiff = (z_2 - z_1) / 2;
  pdf = pdf - 0.5*log(2*3.1415926536);
  pdf = pdf - log( s );
  pdf = pdf - 0.5 * SQR(zDiff) / SQR(s);

  // Now the density of the IG proposal for sig2_2
  pdf = pdf + sigmaConst;
  pdf = pdf - (asigma_ + 1) * log( sig2_2 );
  pdf = pdf - rsigma_ / sig2_2;  

  if(!ln) 
    pdf = exp(pdf);
  return( pdf );
}


/**************************************************************************
 **  Split an atom in the Levy process.  The index of the atom to be
 **  split must be specified in "ind".  The proposal scale for the
 **  magnitude step must be specified in s.  The parameters of the
 **  specified atom are returned in x, sig2, and z.  The parameters
 **  of the new atoms are returned in x_1, sig2_1, z_1, x_2, sig2_2,
 **  and z_2.  The proposal density is returned in pdf, on the log
 **  scale if ln == true; by "proposal density" we mean the density
 **  of the normal proposal for the magnitude step, times the
 **  density of the IG proposal for sig2_2, times the density of the
 **  uniform proposal for x_2.
 **/  
void LevyProc::splitPoint
( int ind, double s, double *x, double *sig2, double *z, 
  double *x_1, double *sig2_1, double *z_1,
  double *x_2, double *sig2_2, double *z_2,   
  double *pdf, bool ln ){

  // get the point to split
  getPoint( ind, x, sig2, z );
  
  (*x_1) = (*x);
  double delta_mu = distAboveAtom( ind );
  (*x_2) = S_runif();
  (*x_2) = (*x_2) * delta_mu;
  (*x_2) = (*x_2) + (*x);
  double step_z = S_rnorm();
  step_z = step_z * s;
  (*z_1) = (*z)/2 - step_z;
  (*z_2) = (*z)/2 + step_z;
  (*sig2_1) = (*sig2);
  // Sample sig2_2 from its inverse gamma xprior
  (*sig2_2) = S_rgamma( asigma_ );
  (*sig2_2) = (*sig2_2) / rsigma_;
  (*sig2_2) = 1 / (*sig2_2);

  (*pdf) = dSplit( ind, (*x), (*z_1), (*sig2_2), (*z_2), s, ln );

}

/**************************************************************************
 **  Merge an atom in the Levy process with the next atom to its
 **  right.  If there is no atom to the right, return -1 (otherwise
 **  return 0).  The index of the atom to be merged must be
 **  specified in "ind1".  The proposal scale (delta_gamma) for the
 **  magnitude step of the opposite split must be specified in s.
 **  The parameters of the specified atom and the next atom to its
 **  right are returned in x_1, sig2_1, z_1 and x_2, sig2_2, z_2,
 **  respectively.  The index of the next atom to the right is
 **  returned in ind2.  The parameters of the new atom are returned
 **  in x, sig2, z.
 **/  
int LevyProc::mergePoints
( int ind1, int *ind2, double s, 
  double *x, double *sig2, double *z, 
  double *x_1, double *sig2_1, double *z_1,
  double *x_2, double *sig2_2, double *z_2 ){

  // get the points to merge
  getPoint( ind1, x_1, sig2_1, z_1 );
  (*ind2) = getPointToRight( ind1, x_2, sig2_2, z_2 );
  // If there is no atom to the right, return -1
  if( (*ind2) < 0 )  return -1;

  (*x) = (*x_1);
  (*sig2) = (*sig2_1);
  (*z) = (*z_1) + (*z_2);

  return 1;
}

/**************************************************************************
 **  Find the next atom to the right of that with index ind1.  Return
 **  the index of that atom, and store its location, variance, and
 **  height in x, sig2, and z.  If there is no atom to the
 **  right, return -1.
 **/  
int LevyProc::getPointToRight
( int ind1, double *x, double *sig2, double *z
){

  // get the location of the point with index ind1.
  double x_1, sig2_1, z_1;
  getPoint( ind1, &x_1, &sig2_1, &z_1 );

  // find the smallest positive location difference
  // and save the index and parameters of that atom.
  double diff, minPosDiff = DBL_MAX;
  int ind2 = -1;
  for( int i=0; i<nAtoms_; i++ ){
    // for each atom
    diff = xLoc_[i] - x_1;
    if( (xLoc_[i] >= x_1) && (diff < minPosDiff) && (i != ind1) ){
      // if it is to the right, and is the closest seen so far,
      // store its index and parameters.
      minPosDiff = diff;
      ind2 = i;
      (*x) = xLoc_[i];
      (*sig2) = atomVar_[i];
      (*z) = atomHeight_[i];
    }
  }

  // If no atoms have been found to the right, then ind2 = -1.
  return ind2;
}

/**************************************************************************
 **  Write the Levy process atom locations and heights out to a
 **  file, appending to the end of the file.
 **/  
void LevyProc::appendToFile( FILE *outfile ){
  double x[1], sig2[1], height[1];
  for (int i=0; i<nAtoms_; i++){
    (*x) = xLoc_[i];
    fwrite( x, sizeof(double), 1, outfile );
    (*sig2) = atomVar_[i];
    fwrite( sig2, sizeof(double), 1, outfile );
    (*height) = atomHeight_[i];
    fwrite( height, sizeof(double), 1, outfile );
  }
}

/**************************************************************************
 **  Print the Levy process atom locations and heights to stdout
 **/  
void LevyProc::print(){
  for (int i=0; i<nAtoms_; i++){
    printf( "%10.10f       %f       %f\n", xLoc_[i],
	    atomVar_[i], atomHeight_[i] );
  }
}

